"""Base class implementation for a package runner"""

# Future
from __future__ import annotations

# Standard Library
import abc
import pathlib
import tempfile
from typing import TYPE_CHECKING, List

# Houdini Package Runner
from houdini_package_runner.config import PackageRunnerConfig

# Imports for type checking.
if TYPE_CHECKING:
    import argparse

    from houdini_package_runner.config import BaseRunnerConfig
    from houdini_package_runner.discoverers.base import BaseItemDiscoverer
    from houdini_package_runner.items.base import BaseItem


# =============================================================================
# CLASSES
# =============================================================================


class HoudiniPackageRunner(abc.ABC):
    """The base package runner class.

    :param discoverer: The item discoverer used by the runner.
    :param write_back: Whether the runner should write the results back.
    :param runner_config: Optional BaseRunnerConfig object.

    """

    def __init__(
        self,
        discoverer: BaseItemDiscoverer,
        write_back: bool = False,
        runner_config: BaseRunnerConfig = None,
    ) -> None:
        self._discoverer = discoverer
        self._extra_args: List[str] = []
        self._hotl_command = "hotl"
        self._list_items = False
        self._temp_dir = pathlib.Path(tempfile.mkdtemp())
        self._verbose = False
        self._write_back = write_back

        if runner_config is None:
            runner_config = PackageRunnerConfig(self.name)

        self._config = runner_config

    # -------------------------------------------------------------------------
    # PROPERTIES
    # -------------------------------------------------------------------------

    @property
    def config(self) -> BaseRunnerConfig:
        """The configuration for the runner."""
        return self._config

    @property
    def discoverer(self) -> BaseItemDiscoverer:
        """The discoverer used by the runner."""
        return self._discoverer

    @property
    def extra_args(self) -> List[str]:
        """A list of extra args to pass to the runner execution."""
        return self._extra_args

    @property
    def hotl_command(self) -> str:
        """The hotl command to use when dealing with digital asset files."""
        return self._hotl_command

    @property
    @abc.abstractmethod
    def name(self) -> str:
        """The runner name used for identification."""

    @property
    def temp_dir(self) -> pathlib.Path:
        """The temp directory used by the package runner."""
        return self._temp_dir

    @property
    def verbose(self) -> bool:
        """Whether to output verbose runner output."""
        return self._verbose

    @property
    def write_back(self) -> bool:
        """Whether the runner should write the results back."""
        return self._write_back

    # -------------------------------------------------------------------------
    # METHODS
    # -------------------------------------------------------------------------

    @abc.abstractmethod
    def process_path(self, file_path: pathlib.Path, item: BaseItem) -> int:
        """Process a file path.

        :param file_path: The path to process.
        :param item: The item to process.
        :return: The process return code.

        """

    def init_args_options(
        self, namespace: argparse.Namespace, extra_args: List[str]
    ):  # pylint: disable=unused-argument
        """Initialize any extra options from parser data.

        :param namespace: Argument parser namespace.
        :param extra_args: Optional list of extra_args to pass to isort.

        """
        self._verbose = namespace.verbose
        self._list_items = namespace.list_items

        if hasattr(namespace, "hotl_command"):
            self._hotl_command = namespace.hotl_command

    def run(self) -> int:
        """Process all the items.

        :return: The overall execution return code.

        """
        if self._list_items:
            for item in self.discoverer.items:
                print(item)

            return 0

        result = 0

        for item in self.discoverer.items:
            if self.write_back:
                item.write_back = True

            result |= item.process(self)

        return result



"""Test the houdini_package_runner.base.item module."""

# =============================================================================
# IMPORTS
# =============================================================================

# Standard Library
import pathlib

# Third Party
import pytest

# Houdini Package Runner
import houdini_package_runner.items.base

# =============================================================================
# FIXTURES
# =============================================================================


@pytest.fixture
def init_base_item(mocker):
    """Initialize a dummy BaseItem for testing."""
    mocker.patch.multiple(
        houdini_package_runner.items.base.BaseItem,
        __abstractmethods__=set(),
        __init__=lambda x, y: None,
    )

    def _create():
        return houdini_package_runner.items.base.BaseItem(None)

    return _create


@pytest.fixture
def init_base_file_item(mocker):
    """Initialize a dummy BaseFileItem for testing."""
    mocker.patch.multiple(
        houdini_package_runner.items.base.BaseFileItem,
        __abstractmethods__=set(),
        __init__=lambda x, y, z: None,
    )

    def _create():
        return houdini_package_runner.items.base.BaseFileItem(None, None)

    return _create


# =============================================================================
# TESTS
# =============================================================================


class TestBaseItem:
    """Test houdini_package_runner.base.item.BaseItem."""

    # Object construction

    @pytest.mark.parametrize("write_back", (None, False, True))
    def test___init__(self, remove_abstract_methods, write_back):
        """Test object initialization."""
        remove_abstract_methods(houdini_package_runner.items.base.BaseItem)

        if write_back is not None:
            inst = houdini_package_runner.items.base.BaseItem(write_back=write_back)

        else:
            inst = houdini_package_runner.items.base.BaseItem()

        assert not inst._contents_changed
        assert isinstance(inst._ignored_builtins, list)
        assert not inst._ignored_builtins
        assert not inst._is_single_line
        assert not inst._is_test_item
        assert inst.write_back == bool(write_back)

    # Properties

    def test_contents_changed(self, init_base_item):
        """Test BaseItem.contents_changed."""
        inst = init_base_item()
        inst._contents_changed = False

        assert not inst.contents_changed

        inst.contents_changed = True
        assert inst._contents_changed

    def test_ignored_builtins(self, mocker, init_base_item):
        """Test BaseItem.ignored_builtins."""
        mock_builtin = mocker.MagicMock(spec=str)
        inst = init_base_item()
        inst._ignored_builtins = [mock_builtin]

        assert inst.ignored_builtins == [mock_builtin]

        with pytest.raises(AttributeError):
            inst.ignored_builtins = []

    def test_is_single_line(self, mocker, init_base_item):
        """Test BaseItem.is_single_line."""
        mock_single = mocker.MagicMock(spec=bool)

        inst = init_base_item()
        inst._is_single_line = mock_single

        assert inst.is_single_line == mock_single

        with pytest.raises(AttributeError):
            inst.is_single_line = True

    def test_is_test_item(self, init_base_item):
        """Test BaseItem.is_test_item."""
        inst = init_base_item()
        inst._is_test_item = False

        assert not inst.is_test_item

        inst.is_test_item = True
        assert inst.is_test_item

    def test_write_back(self, init_base_item):
        """Test BaseItem.write_back."""
        inst = init_base_item()
        inst._write_back = False

        assert not inst.write_back

        inst.write_back = True

        assert inst._write_back


class TestBaseFileItem:
    """Test houdini_package_runner.base.item.BaseFileItem."""

    # Object construction

    @pytest.mark.parametrize("default_args", (False, True))
    def test___init__(self, mocker, remove_abstract_methods, default_args):
        """Test object initialization."""
        mock_path = mocker.MagicMock(spec=pathlib.Path)
        mock_write_back = mocker.MagicMock(spec=bool)

        remove_abstract_methods(houdini_package_runner.items.base.BaseFileItem)

        mock_super_init = mocker.patch.object(
            houdini_package_runner.items.base.BaseItem, "__init__"
        )

        if default_args:
            inst = houdini_package_runner.items.base.BaseFileItem(mock_path)

            mock_super_init.assert_called_with(write_back=False)

        else:
            inst = houdini_package_runner.items.base.BaseFileItem(
                mock_path, write_back=mock_write_back
            )

            mock_super_init.assert_called_with(write_back=mock_write_back)

        assert inst._path == mock_path

    # Properties

    def test_path(self, mocker, init_base_file_item):
        """Test BaseFileItem.path."""
        mock_path = mocker.MagicMock(spec=pathlib.Path)

        inst = init_base_file_item()
        inst._path = mock_path

        assert inst.path == mock_path



"""Test the houdini_package_runner.runners.isort.runner module."""

# =============================================================================
# IMPORTS
# =============================================================================

# Standard Library
import argparse
import configparser
import os
import pathlib

# Third Party
import pytest

# Houdini Package Runner
import houdini_package_runner.config
import houdini_package_runner.items.base
import houdini_package_runner.items.dialog_script
import houdini_package_runner.items.digital_asset
import houdini_package_runner.items.filesystem
import houdini_package_runner.items.xml
import houdini_package_runner.runners.base
import houdini_package_runner.runners.isort.runner
from houdini_package_runner.discoverers.base import BaseItemDiscoverer

# =============================================================================
# FIXTURES
# =============================================================================


@pytest.fixture
def init_runner(mocker):
    """Initialize a dummy IsortRunner for testing."""
    mocker.patch.multiple(
        houdini_package_runner.runners.isort.runner.IsortRunner,
        __init__=lambda x, y: None,
    )

    def _create():
        return houdini_package_runner.runners.isort.runner.IsortRunner(None)

    return _create


# =============================================================================
# TESTS
# =============================================================================


class TestIsortRunner:
    """Test houdini_package_runner.runners.isort.runner.IsortRunner."""

    @pytest.mark.parametrize("pass_optional", (False, True))
    def test___init__(self, mocker, pass_optional):
        """Test object initialization."""
        mock_discoverer = mocker.MagicMock(spec=BaseItemDiscoverer)

        mock_super_init = mocker.patch.object(
            houdini_package_runner.runners.isort.runner.HoudiniPackageRunner, "__init__"
        )

        mock_config = (
            mocker.MagicMock(spec=houdini_package_runner.config.PackageRunnerConfig)
            if pass_optional
            else None
        )

        if pass_optional:
            houdini_package_runner.runners.isort.runner.IsortRunner(
                mock_discoverer, runner_config=mock_config
            )

        else:
            houdini_package_runner.runners.isort.runner.IsortRunner(mock_discoverer)

        mock_super_init.assert_called_with(
            mock_discoverer, write_back=True, runner_config=mock_config
        )

    # Non-Public Methods

    def test__generate_config(self, mocker, init_runner):
        """Test IsortRunner._generate_config."""
        mock_namespace = mocker.MagicMock(spec=argparse.Namespace)

        mock_load = mocker.patch(
            "houdini_package_runner.runners.isort.runner._load_template_config"
        )
        mock_save = mocker.patch(
            "houdini_package_runner.runners.isort.runner._save_template_config"
        )

        mock_temp_dir = mocker.MagicMock(spec=pathlib.Path)

        mock_process = mocker.patch.object(
            houdini_package_runner.runners.isort.runner.IsortRunner, "_process_config"
        )

        mocker.patch.object(
            houdini_package_runner.runners.isort.runner.IsortRunner,
            "temp_dir",
            mock_temp_dir,
        )

        inst = init_runner()

        result = inst._generate_config(mock_namespace)

        assert result == mock_save.return_value

        mock_process.assert_called_with(mock_load.return_value, mock_namespace)
        mock_save.assert_called_with(mock_load.return_value, mock_temp_dir)

    def test__get_first_party_header(self, init_runner):
        """Test IsortRunner._get_first_party_header."""
        package_names = "first_party1,first_party2"

        inst = init_runner()

        result = inst._get_first_party_header(package_names)

        assert result == "First Party1"

    @pytest.mark.parametrize(
        "namespace_packages, python_root_exists",
        (
            (True, True),
            (False, None),
            (False, True),
            (False, False),
        ),
    )
    def test__get_first_party_packages(
        self, mocker, init_runner, namespace_packages, python_root_exists
    ):
        """Test IsortRunner._get_first_party_packages."""

        mock_namespace = mocker.MagicMock(spec=argparse.Namespace)
        mock_namespace.package_names = (
            "first_party1,first_party2" if namespace_packages else None
        )
        mock_namespace.python_root = (
            "python" if python_root_exists is not None else None
        )

        mock_find_python = mocker.patch(
            "houdini_package_runner.runners.isort.runner._find_python_packages_from_path",
            return_value="found_first_party1,found_first_party2",
        )

        mock_discoverer_path = mocker.MagicMock(spec=pathlib.Path)

        mock_discoverer = mocker.MagicMock(spec=BaseItemDiscoverer)
        mock_discoverer.path = mock_discoverer_path
        mocker.patch.object(
            houdini_package_runner.runners.isort.runner.IsortRunner,
            "discoverer",
            mock_discoverer,
        )

        mock_discoverer_path.__truediv__.return_value.exists.return_value = (
            python_root_exists
        )

        inst = init_runner()

        result = inst._get_first_party_packages(mock_namespace)

        if namespace_packages:
            assert result == mock_namespace.package_names

        else:
            if python_root_exists:
                assert result == "found_first_party1,found_first_party2"
                mock_find_python.assert_called_with(mock_discoverer_path / "python")

            else:
                assert result is None
                mock_find_python.assert_not_called()

    def test__get_houdini_names(self, mocker, init_runner):
        """Test IsortRunner._get_houdini_names."""
        mock_namespace = mocker.MagicMock(spec=argparse.Namespace)
        mock_namespace.hfs_path = "$TEMP/houdini19.5"

        mock_find_hfs = mocker.patch(
            "houdini_package_runner.runners.isort.runner._find_known_houdini",
            return_value=["hou", "toolutils"],
        )

        inst = init_runner()

        result = inst._get_houdini_names(mock_namespace)

        assert result == "hou,toolutils"

        mock_find_hfs.assert_called_with(
            pathlib.Path(os.path.expandvars("$TEMP/houdini19.5"))
        )

    @pytest.mark.parametrize("first_party_packages_set", (True, False))
    def test__process_config(self, mocker, init_runner, first_party_packages_set):
        """Test IsortRunner._process_config."""
        settings = {}
        config = {"settings": settings}

        mock_namespace = mocker.MagicMock(spec=argparse.Namespace)

        mock_find_houdini = mocker.patch(
            "houdini_package_runner.runners.isort.runner.IsortRunner._get_houdini_names",
        )
        mock_find_first = mocker.patch(
            "houdini_package_runner.runners.isort.runner.IsortRunner._get_first_party_packages",
        )

        mock_find_first.return_value = (
            "package1,package2" if first_party_packages_set else None
        )

        mock_find_header = mocker.patch(
            "houdini_package_runner.runners.isort.runner.IsortRunner._get_first_party_header",
        )

        inst = init_runner()

        inst._process_config(config, mock_namespace)

        assert settings["known_houdini"] == mock_find_houdini.return_value

        if first_party_packages_set:
            assert settings["known_first_party"] == mock_find_first.return_value
            assert (
                settings["import_heading_firstparty"] == mock_find_header.return_value
            )

        else:
            assert "known_first_party" not in settings

    # Properties

    def test_config_file(self, mocker, init_runner):
        """Test IsortRunner.config_file."""
        mock_config = mocker.MagicMock(spec=str)

        inst = init_runner()
        inst._config_file = mock_config

        assert inst.config_file == mock_config

        inst._config_file = None
        inst.config_file = mock_config
        assert inst._config_file == mock_config

    def test_name(self, init_runner):
        """Test IsortRunner.name."""
        inst = init_runner()

        assert inst.name == "isort"

        with pytest.raises(AttributeError):
            inst.name = []

    # Methods

    @pytest.mark.parametrize("pass_parser", (True, False))
    def test_build_parser(self, mocker, pass_parser):
        """Test IsortRunner.build_parser."""
        mock_parser = mocker.MagicMock(spec=argparse.ArgumentParser)

        mock_build = mocker.patch(
            "houdini_package_runner.parser.build_common_parser",
            return_value=mock_parser,
        )

        if pass_parser:
            result = (
                houdini_package_runner.runners.isort.runner.IsortRunner.build_parser(
                    parser=mock_parser
                )
            )

            mock_build.assert_not_called()

        else:
            result = (
                houdini_package_runner.runners.isort.runner.IsortRunner.build_parser()
            )

            mock_build.assert_called()

        assert result == mock_parser

        result.add_argument.assert_has_calls(
            [
                mocker.call(
                    "--config-file",
                    action="store",
                    default=".isort.cfg",
                    help="Optional config file to pass to isort commands",
                ),
                mocker.call(
                    "--hfs-path",
                    action="store",
                    default="$HFS",
                    help="Path to a Houdini install directory for known Houdini modules",
                ),
                mocker.call(
                    "--package-names",
                    action="store",
                    help="Known first party package names",
                ),
            ]
        )

    @pytest.mark.parametrize(
        "namespace_config_exists",
        (None, True, False),
    )
    def test_init_args_options(self, mocker, init_runner, namespace_config_exists):
        """Test IsortRunner.init_args_options."""
        mock_namespace = mocker.MagicMock(spec=argparse.Namespace)
        mock_namespace.config_file = None

        mock_discoverer_path = mocker.MagicMock(spec=pathlib.Path)

        mock_discoverer = mocker.MagicMock(spec=BaseItemDiscoverer)
        mock_discoverer.path = mock_discoverer_path
        mocker.patch.object(
            houdini_package_runner.runners.isort.runner.IsortRunner,
            "discoverer",
            mock_discoverer,
        )

        mock_config_file = mocker.MagicMock(spec=str)

        if namespace_config_exists is not None:
            mock_namespace.config_file = mock_config_file
            mock_discoverer.path.__truediv__.return_value.exists.return_value = (
                namespace_config_exists
            )

        mock_extra_args = mocker.MagicMock(spec=list)

        mock_super_init = mocker.patch.object(
            houdini_package_runner.runners.isort.runner.HoudiniPackageRunner,
            "init_args_options",
        )

        mock_generate = mocker.patch.object(
            houdini_package_runner.runners.isort.runner.IsortRunner, "_generate_config"
        )

        inst = init_runner()
        inst._config_file = None

        inst.init_args_options(mock_namespace, mock_extra_args)

        mock_super_init.assert_called_with(mock_namespace, mock_extra_args)

        assert inst._extra_args == mock_extra_args

        if namespace_config_exists is not None:
            mock_discoverer.path.__truediv__.assert_called_with(mock_config_file)

        if namespace_config_exists:
            assert inst.config_file == mock_discoverer.path.__truediv__.return_value

        else:
            assert inst.config_file == mock_generate.return_value

            mock_generate.assert_called_with(mock_namespace)

    @pytest.mark.parametrize(
        "has_config",
        (True, False),
    )
    def test_process_path(self, mocker, init_runner, has_config):
        """Test IsortRunner.process_path."""
        mock_path = mocker.MagicMock(spec=pathlib.Path)

        mock_item = mocker.MagicMock(
            spec=houdini_package_runner.items.filesystem.FileToProcess
        )

        mock_execute = mocker.patch(
            "houdini_package_runner.utils.execute_subprocess_command"
        )

        extra_args = ["--flag", "arg"]

        mocker.patch.object(
            houdini_package_runner.runners.isort.runner.IsortRunner,
            "extra_args",
            extra_args,
        )

        mock_config_path = mocker.MagicMock(spec=pathlib.Path)
        config_file = mock_config_path if has_config else None

        mocker.patch.object(
            houdini_package_runner.runners.isort.runner.IsortRunner,
            "config_file",
            config_file,
        )

        mock_verbose = mocker.MagicMock(spec=bool)

        inst = init_runner()
        inst._verbose = mock_verbose

        inst.process_path(mock_path, mock_item)

        expected_args = ["isort"]

        if has_config:
            expected_args.extend(["--sp", str(mock_config_path)])

        expected_args.extend(extra_args + [str(mock_path)])

        mock_execute.assert_called_with(expected_args, verbose=mock_verbose)


def test__find_known_houdini(shared_datadir):
    """Test houdini_package_runner.runners.isort.runner._find_known_houdini."""
    test_path = shared_datadir / "test__find_known_houdini"
    result = houdini_package_runner.runners.isort.runner._find_known_houdini(test_path)

    assert result == ["IFDframe", "hjson", "hou", "hutil"]


def test__find_python_modules(shared_datadir):
    """Test houdini_package_runner.runners.isort.runner._find_python_modules."""
    test_path = shared_datadir / "test__find_python_modules"

    result = houdini_package_runner.runners.isort.runner._find_python_modules(test_path)

    assert result == ["compiled_module", "python_file", "test_dir"]


@pytest.mark.parametrize("packages_found", (True, False))
def test__find_python_packages_from_path(shared_datadir, packages_found):
    """Test houdini_package_runner.runners.isort.runner._find_python_packages_from_path."""
    if packages_found:
        test_path = (
            shared_datadir / "test__find_python_packages_from_path" / "has_packages"
        )

    else:
        test_path = shared_datadir / "test__find_python_packages_from_path"

    result = (
        houdini_package_runner.runners.isort.runner._find_python_packages_from_path(
            test_path
        )
    )

    if packages_found:
        assert result == "test_other_package,test_package"

    else:
        assert result is None


def test__load_template_config(mocker):
    """Test houdini_package_runner.runners.isort.runner._load_template_config."""
    mock_open_text = mocker.patch("importlib.resources.open_text")

    mock_config = mocker.patch(
        "houdini_package_runner.runners.isort.runner.ConfigParser"
    )

    result = houdini_package_runner.runners.isort.runner._load_template_config()

    assert result == mock_config.return_value

    mock_config.return_value.read_file.assert_called_with(
        mock_open_text.return_value.__enter__.return_value
    )


def test__save_template_config(mocker):
    """Test houdini_package_runner.runners.isort.runner._save_template_config."""
    mock_config = mocker.MagicMock(spec=configparser.ConfigParser)
    mock_temp_path = mocker.MagicMock(spec=pathlib.Path)

    mock_handle = mocker.mock_open()
    mocker.patch("builtins.open", mock_handle)

    result = houdini_package_runner.runners.isort.runner._save_template_config(
        mock_config, mock_temp_path
    )

    assert result == mock_temp_path / ".isort.cfg"

    mock_handle.assert_called_with(mock_temp_path / ".isort.cfg", "w", encoding="utf-8")

    mock_config.write.assert_called_with(mock_handle.return_value)


def test_main(mocker):
    """Test houdini_package_runner.runners.isort.runner.main."""
    mock_parsed = mocker.MagicMock(spec=argparse.Namespace)
    mock_unknown = mocker.MagicMock(spec=list)

    mock_parser = mocker.MagicMock(spec=argparse.ArgumentParser)
    mock_parser.parse_known_args.return_value = (mock_parsed, mock_unknown)

    mock_discoverer = mocker.MagicMock(spec=BaseItemDiscoverer)
    mock_init = mocker.patch(
        "houdini_package_runner.runners.flake8.runner.package.init_standard_package_discoverer",
        return_value=mock_discoverer,
    )

    mock_runner = mocker.MagicMock(
        spec=houdini_package_runner.runners.isort.runner.IsortRunner
    )

    mock_runner_init = mocker.patch(
        "houdini_package_runner.runners.isort.runner.IsortRunner",
        return_value=mock_runner,
    )
    mock_runner_init.build_parser.return_value = mock_parser

    result = houdini_package_runner.runners.isort.runner.main()
    assert result == mock_runner.run.return_value

    mock_init.assert_called_with(mock_parsed)

    mock_runner_init.assert_called_with(mock_discoverer)
    mock_runner.init_args_options.assert_called_with(mock_parsed, mock_unknown)
    mock_runner.run.assert_called()



"""This module contains utility functions."""

# =============================================================================
# IMPORTS
# =============================================================================

# Standard Library
import os
import subprocess
from typing import List

# =============================================================================
# FUNCTIONS
# =============================================================================


def add_or_append_to_flags(
    flags: List[str], key: str, values: List[str], separator: str = ","
):
    """Add or append to an existing arg flag with the separator.

    :param flags: The list of flags to modify.
    :param key: The new flag.
    :param values: The flag options.
    :param separator: The character used to join flag values.
    :return:

    """
    flag_str = separator.join(values)

    if key in flags:
        idx = flags.index(key) + 1
        flags[idx] = flags[idx] + separator + flag_str

    else:
        flags.extend((key, flag_str))


def execute_subprocess_command(command: List[str], verbose: bool = False) -> int:
    """Execute a command in a subprocess, capturing and optionally outputting the output.

    :param command: A subprocess command list.
    :param verbose: Whether to capture output.
    :return: The subprocess return code.

    """
    # If verbose mode is turned on then we don't want to capture the output.
    # so we'll set the output pipe to be None, otherwise we'll capture it.
    output_pipe = None if verbose else subprocess.PIPE

    env = os.environ.copy()

    # Remove PYTHONHOME from the env if it exists. This can cause problems for subprocesses
    # being executed from within Houdini.
    if "PYTHONHOME" in env:
        del env["PYTHONHOME"]

    with subprocess.Popen(
        command, stdout=output_pipe, stderr=output_pipe, env=env
    ) as proc:
        proc.wait()

        return_code = proc.returncode

        if return_code and not verbose:
            if proc.stdout is not None:  # pragma: no branch
                for line in proc.stdout.readlines():
                    print(line.decode("utf-8"))

            if proc.stderr is not None:  # pragma: no branch
                for line in proc.stderr.readlines():
                    print(line.decode("utf-8"))

        return return_code



"""Test the houdini_package_runner.discoverers.base module."""

# =============================================================================
# IMPORTS
# =============================================================================

# Standard Library
import pathlib

# Third Party
import pytest

# Houdini Package Runner
import houdini_package_runner.discoverers.base
import houdini_package_runner.items.filesystem

# =============================================================================
# FIXTURES
# =============================================================================


@pytest.fixture
def init_discoverer(mocker):
    """Initialize a dummy HoudiniPackageRunner for testing."""
    mocker.patch.multiple(
        houdini_package_runner.discoverers.base.BaseItemDiscoverer,
        __abstractmethods__=set(),
        __init__=lambda x, y, z: None,
    )

    def _create():
        return houdini_package_runner.discoverers.base.BaseItemDiscoverer(None, None)

    return _create


# =============================================================================
# TESTS
# =============================================================================


class TestBaseItemDiscoverer:
    """Test houdini_package_runner.discoverers.base.BaseItemDiscoverer."""

    # Object construction

    @pytest.mark.parametrize("has_items", (False, True))
    def test___init__(self, mocker, remove_abstract_methods, has_items):
        """Test object initialization."""
        remove_abstract_methods(
            houdini_package_runner.discoverers.base.BaseItemDiscoverer
        )
        mock_path = mocker.MagicMock(spec=pathlib.Path)

        items = []

        if has_items:
            items = [
                mocker.MagicMock(
                    spec=houdini_package_runner.items.filesystem.FileToProcess
                )
            ]
            inst = houdini_package_runner.discoverers.base.BaseItemDiscoverer(
                mock_path, items=items
            )

        else:
            inst = houdini_package_runner.discoverers.base.BaseItemDiscoverer(mock_path)

        assert inst._path == mock_path
        assert inst._items == items

    # Properties

    def test_items(self, mocker, init_discoverer):
        """Test BaseItemDiscoverer.items."""
        mock_items = mocker.MagicMock(spec=list)

        inst = init_discoverer()
        inst._items = mock_items

        assert inst.items == mock_items

    def test_temp_dir(self, mocker, init_discoverer):
        """Test BaseItemDiscoverer.temp_dir."""
        mock_path = mocker.MagicMock(spec=pathlib.Path)

        inst = init_discoverer()
        inst._path = mock_path

        assert inst.path == mock_path



"""This module contains a class to discover package items."""

# =============================================================================
# IMPORTS
# =============================================================================

# Future
from __future__ import annotations

# Standard Library
import argparse
import pathlib
from typing import List, Tuple

# =============================================================================
# CLASSES
# =============================================================================


class _UltimateHelpFormatter(
    argparse.RawTextHelpFormatter, argparse.ArgumentDefaultsHelpFormatter
):
    """Formatter class that combines RawTextHelpFormatter and ArgumentDefaultsHelpFormatter."""


# =============================================================================
# NON-PUBLIC FUNCTIONS
# =============================================================================


def _resolve_houdini_root(
    parsed_args: argparse.Namespace, root: pathlib.Path
) -> pathlib.Path:
    """Resolve the Houdini root from the args and path.

    :param parsed_args: The parsed command args.
    :param root: The root path.
    :return: The resolved Houdini root.

    """
    if parsed_args.houdini_root is not None:
        houdini_root_value = parsed_args.houdini_root

        if pathlib.Path(houdini_root_value).is_dir():
            return pathlib.Path(houdini_root_value)

        if (root / houdini_root_value).is_dir():
            return root / houdini_root_value

        raise OSError("Could not find houdini root")

    if (root / "houdini").is_dir():
        houdini_root = root / "houdini"

    else:
        houdini_root = root

    return houdini_root


def _resolve_python_packages(
    parsed_args: argparse.Namespace, root: pathlib.Path
) -> List[pathlib.Path]:
    """Find any Python package paths from the args and path.

    :param parsed_args: The parsed command args.
    :param root: The root path.
    :return: A list of any found Python package directories.

    """
    if parsed_args.python_root:
        python_dir = root / parsed_args.python_root

        if python_dir.is_dir():
            return [python_dir]

    return []


def _resolve_tests(
    parsed_args: argparse.Namespace, root: pathlib.Path
) -> List[pathlib.Path]:
    """Find any test paths from the args and path.

    :param parsed_args: The parsed command args.
    :param root: The root path.
    :return: A list of any test directories.

    """
    directories = []

    if not parsed_args.skip_tests:
        test_dir = root / "tests"

        if test_dir.is_dir():
            directories.append(test_dir)

    return directories


# =============================================================================
# FUNCTIONS
# =============================================================================


def build_common_parser(
    prog: str = None, usage: str = None, description: str = None
) -> argparse.ArgumentParser:
    """Build an argument parser for the script args.

    :param prog: The name of the program (default: sys.argv[0])
    :param usage: The string describing the program usage (default: generated from arguments added to parser)
    :param description: Text to display before the argument help.
    :return: The constructed parser.

    """
    parser = argparse.ArgumentParser(
        prog=prog,
        usage=usage,
        description=description,
        formatter_class=_UltimateHelpFormatter,
    )

    parser.add_argument(
        "--add-dir",
        action="append",
        dest="directories",
        default=[],
        help="Add a directory to be processed",
    )

    parser.add_argument(
        "--add-file",
        action="append",
        dest="files",
        default=[],
        help="Add a file to be processed",
    )

    parser.add_argument(
        "--houdini-items",
        action="store",
        default="otls,hda,toolbar,python_panels,pythonXlibs,scripts,soho,menus",
        help="A list of Houdini items to process",
    )

    parser.add_argument(
        "--houdini-root",
        action="store",
        help="The root directory that the Houdini items will be in",
    )

    parser.add_argument(
        "--python-root",
        action="store",
        default="python",
        help="The root directory the package will be in",
    )

    parser.add_argument(
        "--root",
        action="store",
        help="Optional root directory to look for things from.  By default uses the cwd",
    )

    parser.add_argument("--verbose", action="store_true", help="Engage verbose output.")

    parser.add_argument("--hotl-command", action="store", default="hotl")

    parser.add_argument(
        "--skip-tests",
        action="store_true",
        help="Skip processing of files under {root}/tests.",
    )

    parser.add_argument(
        "--list-items",
        action="store_true",
        help="Only list the found items and do not execute the runner.",
    )

    return parser


def process_common_arg_settings(
    parsed_args: argparse.Namespace,
) -> Tuple[pathlib.Path, pathlib.Path, List[pathlib.Path], List[str]]:
    """Generate base settings from common command args.

    :param parsed_args: The parsed command args.
    :return: The root path, houdini root path, directories and files to process, and Houdini item names.

    """
    if parsed_args.root:
        root = pathlib.Path(parsed_args.root)

    else:
        root = pathlib.Path.cwd()

    houdini_root = _resolve_houdini_root(parsed_args, root)

    extra_paths = [root / dir_name for dir_name in parsed_args.directories]

    extra_paths.extend(_resolve_python_packages(parsed_args, root))

    extra_paths.extend(_resolve_tests(parsed_args, root))

    extra_paths.extend([root / file_name for file_name in parsed_args.files])

    houdini_items = parsed_args.houdini_items.split(",")

    return root, houdini_root, extra_paths, houdini_items



"""This module contains a class to discover package items."""

# =============================================================================
# IMPORTS
# =============================================================================

# Future
from __future__ import annotations

# Standard Library
import abc
import pathlib
from typing import TYPE_CHECKING, List, Optional

# Imports for type checking.
if TYPE_CHECKING:
    from houdini_package_runner.items.base import BaseItem


# =============================================================================
# CLASSES
# =============================================================================


class BaseItemDiscoverer(metaclass=abc.ABCMeta):
    """This class is responsible for searching for various items to process.

    :param path: The path to start searching for items from.
    :param items: Optional list of specific items to check.

    """

    def __init__(
        self,
        path: pathlib.Path,
        items: Optional[List[BaseItem]] = None,
    ) -> None:
        self._items = [] if items is None else items
        self._path = path

    # -------------------------------------------------------------------------
    # PROPERTIES
    # -------------------------------------------------------------------------

    @property
    def items(self) -> List[BaseItem]:
        """The list of found items for processing."""
        return self._items

    @property
    def path(self) -> pathlib.Path:
        """The root path."""
        return self._path

    # -------------------------------------------------------------------------
    # METHODS
    # -------------------------------------------------------------------------

    def post_run(self) -> None:
        """Optionally perform any operations after running."""



"""Test the houdini_package_runner.config module."""

# =============================================================================
# IMPORTS
# =============================================================================

# Standard Library
import importlib.resources
import os
import pathlib

# Third Party
import pytest
import toml

# Houdini Package Runner
import houdini_package_runner.config
import houdini_package_runner.items.base

# =============================================================================
# FIXTURES
# =============================================================================


@pytest.fixture
def init_base_config(mocker):
    """Initialize a dummy BaseRunnerConfig for testing."""
    mocker.patch.multiple(
        houdini_package_runner.config.BaseRunnerConfig,
        __abstractmethods__=set(),
        __init__=lambda x, y: None,
    )

    def _create():
        return houdini_package_runner.config.BaseRunnerConfig(None)

    return _create


@pytest.fixture
def init_package_config(mocker):
    """Initialize a dummy PackageRunnerConfig for testing."""
    mocker.patch.multiple(
        houdini_package_runner.config.BaseRunnerConfig,
        __init__=lambda x, y: None,
    )

    def _create():
        return houdini_package_runner.config.PackageRunnerConfig(None)

    return _create


# =============================================================================
# TESTS
# =============================================================================


class TestBaseRunnerConfig:
    """Test houdini_package_runner.config.BaseRunnerConfig."""

    def test___init__(self, mocker, remove_abstract_methods):
        """Test object initialization."""
        remove_abstract_methods(houdini_package_runner.config.BaseRunnerConfig)

        mock_load = mocker.patch(
            "houdini_package_runner.config.BaseRunnerConfig.load_config"
        )

        mock_name = mocker.MagicMock(spec=str)

        inst = houdini_package_runner.config.BaseRunnerConfig(mock_name)

        assert inst._runner_name == mock_name
        assert inst._data == mock_load.return_value

    # Properties

    def test_data(self, mocker, init_base_config):
        """Test BaseRunnerConfig.data"""
        mock_data = mocker.MagicMock(spec=dict)

        inst = init_base_config()
        inst._data = mock_data

        assert inst.data == mock_data

        with pytest.raises(AttributeError):
            inst.data = {}

    def test_runner_name(self, mocker, init_base_config):
        """Test BaseRunnerConfig.runner_name"""
        mock_name = mocker.MagicMock(spec=str)

        inst = init_base_config()
        inst._runner_name = mock_name

        assert inst.runner_name == mock_name

        with pytest.raises(AttributeError):
            inst.runner_name = None


class TestPackageRunnerConfig:
    """Test houdini_package_runner.config.PackageRunnerConfig."""

    def test__get_file_config_data(self, mocker, shared_datadir, init_package_config):
        """Test PackageRunnerConfig._get_file_config_data"""
        config_file = shared_datadir / "get_file_config_data.toml"

        with config_file.open() as handle:
            data = toml.load(handle)

        file_path = pathlib.Path("/foo/OnCreated.py")

        inst = init_package_config()
        inst._data = data

        result = inst._get_file_config_data(file_path, "to_ignore")

        assert result == ["W292"]

    @pytest.mark.parametrize("test_item", (True, False))
    def test__get_item_config_data(
        self, mocker, shared_datadir, init_package_config, test_item
    ):
        """Test PackageRunnerConfig._get_item_config_data"""
        config_file = shared_datadir / "get_item_config_data.toml"

        with config_file.open() as handle:
            data = toml.load(handle)

        mock_build = mocker.patch(
            "houdini_package_runner.config.build_config_item_list",
            return_value=("XMLBase", "Foo", "BaseItem"),
        )

        mock_item = mocker.MagicMock(spec=houdini_package_runner.items.base.BaseItem)
        mock_item.is_test_item = test_item

        inst = init_package_config()
        inst._data = data

        result = inst._get_item_config_data(mock_item, "to_ignore")

        expected = ["W292", "A123"]

        if test_item:
            expected.append("T789")

        assert result == expected

        mock_build.assert_called_with(mock_item)

    def test_get_config_data(self, mocker, init_package_config):
        """Test PackageRunnerConfig.get_config_data"""
        mock_key = mocker.MagicMock(spec=str)
        mock_item = mocker.MagicMock(spec=houdini_package_runner.items.base.BaseItem)
        mock_path = mocker.MagicMock(spec=pathlib.Path)

        mock_get_item_data = mocker.patch.object(
            houdini_package_runner.config.PackageRunnerConfig,
            "_get_item_config_data",
            return_value=["item", "result"],
        )
        mock_get_file_data = mocker.patch.object(
            houdini_package_runner.config.PackageRunnerConfig,
            "_get_file_config_data",
            return_value=["file", "result"],
        )

        inst = init_package_config()

        result = inst.get_config_data(mock_key, mock_item, mock_path)

        assert result == ["item", "result", "file", "result"]

        mock_get_item_data.assert_called_with(mock_item, mock_key)
        mock_get_file_data.assert_called_with(mock_path, mock_key)

    def test_load_config(self, mocker, init_package_config):
        """Test PackageRunnerConfig.load_config"""
        mock_load = mocker.patch(
            "houdini_package_runner.config._load_default_runner_config"
        )
        mock_name = mocker.MagicMock(spec=str)

        inst = init_package_config()
        inst._runner_name = mock_name

        result = inst.load_config()

        assert result == mock_load.return_value

        mock_load.assert_called_with(mock_name)


@pytest.mark.parametrize("config_path_set", (True, False))
def test__find_config_files(mocker, config_path_set):
    """Test houdini_package_runner.config._find_config_files."""
    env_dict = {}

    if config_path_set:
        env_dict["HOUDINI_PACKAGE_RUNNER_CONFIG_PATH"] = os.pathsep.join(
            ["path1", "path2", "path3"]
        )
        mocker.patch("os.path.exists", side_effect=(True, False, True))

    mocker.patch.dict(os.environ, env_dict, clear=True)

    result = houdini_package_runner.config._find_config_files()

    if config_path_set:
        assert result == [pathlib.Path("path1"), pathlib.Path("path3")]

    else:
        with importlib.resources.path("houdini_package_runner", "runners.toml") as path:
            assert result == [path]


def test__load_default_runner_config(mocker, shared_datadir):
    """Test houdini_package_runner.config._load_default_runner_config."""
    test_data_path = shared_datadir / "test_load_runner_config"

    mocker.patch(
        "houdini_package_runner.config._find_config_files",
        return_value=[test_data_path / "config1.toml", test_data_path / "config2.toml"],
    )

    expected_path = test_data_path / "expected.toml"

    with expected_path.open("r") as handle:
        expected_data = toml.load(handle)

    result = houdini_package_runner.config._load_default_runner_config("runner_name")

    assert result == expected_data.get("runner_name")

    result = houdini_package_runner.config._load_default_runner_config("none")

    assert result == {}


def test_build_config_item_list():
    """Test houdini_package_runner.config.build_config_item_list.

    This will also test the recursive function __get_base_classes.

    """

    class A:
        """Test base class."""

    class B(A):
        """Another class"""

    class C(B):
        """Child class"""

    inst = C()

    result = houdini_package_runner.config.build_config_item_list(inst)

    assert result == ("C", "B", "A")



"""Test the houdini_package_runner.items.digital_asset module."""

# =============================================================================
# IMPORTS
# =============================================================================

# Standard Library
import pathlib

# Third Party
import pytest
from lxml import etree

# Houdini Package Runner
import houdini_package_runner.items.base
import houdini_package_runner.items.filesystem
import houdini_package_runner.items.xml
import houdini_package_runner.runners.base

# =============================================================================
# FIXTURES
# =============================================================================


@pytest.fixture
def init_base(mocker):
    """Initialize a dummy XMLBase for testing."""
    mocker.patch.multiple(
        houdini_package_runner.items.xml.XMLBase,
        __abstractmethods__=set(),
        __init__=lambda x, y, z, u: None,
    )

    def _create():
        return houdini_package_runner.items.xml.XMLBase(None, None, None)

    return _create


@pytest.fixture
def init_menu(mocker):
    """Initialize a dummy MenuFile for testing."""
    mocker.patch.multiple(
        houdini_package_runner.items.xml.MenuFile,
        __init__=lambda x, y, z, u: None,
    )

    def _create():
        return houdini_package_runner.items.xml.MenuFile(None, None, None)

    return _create


@pytest.fixture
def init_panel(mocker):
    """Initialize a dummy PythonPanelFile for testing."""
    mocker.patch.multiple(
        houdini_package_runner.items.xml.PythonPanelFile,
        __init__=lambda x, y, z, u: None,
    )

    def _create():
        return houdini_package_runner.items.xml.PythonPanelFile(None, None, None)

    return _create


@pytest.fixture
def init_shelf(mocker):
    """Initialize a dummy ShelfFile for testing."""
    mocker.patch.multiple(
        houdini_package_runner.items.xml.ShelfFile,
        __init__=lambda x, y, z, u, v: None,
    )

    def _create():
        return houdini_package_runner.items.xml.ShelfFile(None, None, None, None)

    return _create


# =============================================================================
# TESTS
# =============================================================================


class TestXMLBase:
    """Test houdini_package_runner.items.xml.XMLBase."""

    @pytest.mark.parametrize("default_args", (False, True))
    def test___init__(self, mocker, remove_abstract_methods, default_args):
        """Test object initialization."""
        remove_abstract_methods(houdini_package_runner.items.xml.XMLBase)

        mock_path = mocker.MagicMock(spec=pathlib.Path)
        mock_write_back = mocker.MagicMock(spec=bool)
        mock_display_name = mocker.MagicMock(spec=str)

        if default_args:
            inst = houdini_package_runner.items.xml.XMLBase(mock_path)

            assert not inst.write_back
            assert inst.display_name is None

        else:
            inst = houdini_package_runner.items.xml.XMLBase(
                mock_path, write_back=mock_write_back, display_name=mock_display_name
            )

            assert inst.write_back == mock_write_back
            assert inst.display_name == mock_display_name

        assert "hou" in inst.ignored_builtins
        assert "kwargs" in inst.ignored_builtins

    @pytest.mark.parametrize(
        "has_cdata, contents_changed",
        (
            (True, True),
            (True, False),
            (False, True),
            (False, False),
        ),
    )
    def test__handle_write_back(self, mocker, init_base, has_cdata, contents_changed):
        """Test XMLBase._handle_write_back."""
        tostring_result = (
            "something CDATA something" if has_cdata else "something something"
        )

        mock_tostring = mocker.patch(
            "houdini_package_runner.items.xml.etree.tostring",
            return_value=tostring_result,
        )
        mock_cdata = mocker.patch("houdini_package_runner.items.xml.etree.CDATA")

        mock_text = mocker.MagicMock(spec=str)
        mock_changed_text = mocker.MagicMock(spec=str)

        mock_cdata.return_value = mock_changed_text if contents_changed else mock_text

        mock_section = mocker.MagicMock()
        mock_section.text = mock_text

        mock_temp_path = mocker.MagicMock(spec=pathlib.Path)

        mock_handle = mocker.mock_open()
        mock_handle.return_value.read.return_value = (
            mock_changed_text if contents_changed else mock_text
        )
        mock_temp_path.open = mock_handle

        inst = init_base()
        inst._contents_changed = False

        inst._handle_write_back(mock_section, mock_temp_path)

        mock_tostring.assert_called_with(mock_section, encoding="unicode")

        if contents_changed:
            assert inst.contents_changed

            if has_cdata:
                assert mock_section.text == etree.CDATA(mock_changed_text)

            else:
                assert mock_section.text == mock_changed_text

        else:
            if has_cdata:
                assert mock_section.text == etree.CDATA(mock_text)

            else:
                assert mock_section.text == mock_text

    @pytest.mark.parametrize("write_back", (True, False))
    def test__process_code_section(self, mocker, init_base, write_back):
        """Test XMLBase._process_code_section."""
        mock_text = mocker.MagicMock(spec=str)

        mock_section = mocker.MagicMock()
        mock_section.text = mock_text

        mock_base_name = mocker.MagicMock(spec=str)

        mock_handle = mocker.mock_open()

        mock_temp_path = mocker.MagicMock(spec=pathlib.Path)
        mock_temp_path.open = mock_handle

        mock_runner = mocker.MagicMock(
            spec=houdini_package_runner.runners.base.HoudiniPackageRunner
        )
        mock_runner.temp_dir.__truediv__.return_value = mock_temp_path

        mock_handle_write_back = mocker.patch.object(
            houdini_package_runner.items.xml.XMLBase, "_handle_write_back"
        )

        inst = init_base()
        inst.write_back = write_back

        result = inst._process_code_section(mock_section, mock_runner, mock_base_name)

        assert result == mock_runner.process_path.return_value

        mock_runner.temp_dir.__truediv__.assert_called_with(f"{mock_base_name}.py")

        mock_temp_path.open.assert_called_with("w")
        mock_handle.return_value.write.assert_called_with(str(mock_text))

        mock_runner.process_path.assert_called_with(mock_temp_path, inst)

        if write_back:
            mock_handle_write_back.assert_called_with(mock_section, mock_temp_path)

        else:
            mock_handle_write_back.assert_not_called()

    @pytest.mark.parametrize(
        "write_back, contents_changed, return_codes, expected",
        (
            (True, True, (0, 0), 0),
            (True, False, (1, 0), 1),
            (False, True, (0, 1), 1),
            (False, False, (0, 0), 0),
        ),
    )
    def test_process(
        self, mocker, init_base, write_back, contents_changed, return_codes, expected
    ):
        """Test XMLBase.process."""
        mock_process = mocker.patch.object(
            houdini_package_runner.items.xml.XMLBase,
            "_process_code_section",
            side_effect=return_codes,
        )

        mock_section1 = mocker.MagicMock()
        mock_name1 = mocker.MagicMock()

        mock_section2 = mocker.MagicMock()
        mock_name2 = mocker.MagicMock()

        mock_get_items = mocker.patch.object(
            houdini_package_runner.items.xml.XMLBase, "_get_items_to_process"
        )
        mock_get_items.return_value = (
            (mock_section1, mock_name1),
            (mock_section2, mock_name2),
        )

        mock_root = mocker.MagicMock()

        mock_tree = mocker.MagicMock()
        mock_tree.getroot.return_value = mock_root

        mock_parser = mocker.patch("houdini_package_runner.items.xml.etree.XMLParser")
        mock_parse = mocker.patch(
            "houdini_package_runner.items.xml.etree.parse", return_value=mock_tree
        )

        mock_runner = mocker.MagicMock(
            spec=houdini_package_runner.runners.base.HoudiniPackageRunner
        )

        mock_path = mocker.MagicMock(spec=pathlib.Path)

        inst = init_base()
        inst._path = mock_path
        inst.write_back = write_back
        inst.contents_changed = contents_changed

        result = inst.process(mock_runner)

        assert result == expected

        mock_parser.assert_called_with(strip_cdata=False)
        mock_parse.assert_called_with(str(inst.path), mock_parser.return_value)
        mock_get_items.assert_called_with(mock_root)

        mock_process.assert_has_calls(
            (
                mocker.call(mock_section1, mock_runner, mock_name1),
                mocker.call(mock_section2, mock_runner, mock_name2),
            )
        )

        if write_back and contents_changed:
            mock_tree.write.assert_called_with(
                str(inst.path), encoding="utf-8", xml_declaration=True
            )

        else:
            mock_tree.write.assert_not_called()


class TestMenuFile:
    """Test houdini_package_runner.items.xml.MenuFile."""

    def test__get_items_to_process(self, shared_datadir, init_menu):
        """Test MenuFile._get_items_to_process."""
        test_path = shared_datadir / "test_menu__get_items_to_process.xml"
        root = _load_test_file(test_path)

        inst = init_menu()

        result = inst._get_items_to_process(root)

        assert len(result) == 4

        assert result[0][0].text == "promote code"
        assert result[0][1] == "promote_item"

        assert result[1][0].text == "item code"
        assert result[1][1] == "item_with_context"

        assert result[2][0].text == "context code"
        assert result[2][1] == "item_with_context.context"

        assert result[3][0].text == "missing context item code"
        assert result[3][1] == "item_with_context_missing_expressions"


class TestPythonPanelFile:
    """Test houdini_package_runner.items.xml.PythonPanelFile."""

    def test__get_items_to_process(self, shared_datadir, init_panel):
        """Test PythonPanelFile._get_items_to_process."""
        test_path = shared_datadir / "test_pypanel__get_items_to_process.xml"
        root = _load_test_file(test_path)

        inst = init_panel()

        result = inst._get_items_to_process(root)

        assert len(result) == 2

        assert result[0][0].text == "script code"
        assert result[0][1] == "test_panel"

        assert result[1][0].text == "more script code"
        assert result[1][1] == "test_other_panel"


class TestShelfFile:
    """Test houdini_package_runner.items.xml.ShelfFile."""

    @pytest.mark.parametrize("default_args", (False, True))
    def test___init__(self, mocker, default_args):
        """Test object initialization."""
        mock_path = mocker.MagicMock(spec=pathlib.Path)
        mock_write_back = mocker.MagicMock(spec=bool)
        mock_display_name = mocker.MagicMock(spec=str)
        mock_tool_name = mocker.MagicMock(spec=str)

        mock_super_init = mocker.patch.object(
            houdini_package_runner.items.xml.XMLBase, "__init__"
        )

        if default_args:
            inst = houdini_package_runner.items.xml.ShelfFile(mock_path)

            mock_super_init.assert_called_with(
                mock_path, write_back=False, display_name=None
            )

            assert inst._tool_name is None

        else:
            inst = houdini_package_runner.items.xml.ShelfFile(
                mock_path,
                write_back=mock_write_back,
                display_name=mock_display_name,
                tool_name=mock_tool_name,
            )

            mock_super_init.assert_called_with(
                mock_path, write_back=mock_write_back, display_name=mock_display_name
            )

            assert inst._tool_name == mock_tool_name

    @pytest.mark.parametrize("tool_name", (None, "com.houdinitoolbox::Sop/foo::1"))
    def test__get_items_to_process(self, shared_datadir, init_shelf, tool_name):
        """Test ShelfFile._get_items_to_process."""
        test_path = shared_datadir / "test_shelf__get_items_to_process.shelf"
        root = _load_test_file(test_path)

        inst = init_shelf()
        inst._tool_name = tool_name

        result = inst._get_items_to_process(root)

        assert len(result) == 2

        assert result[0][0].text == "shelf code"
        assert result[0][1] == "copy_items"

        assert result[1][0].text == "hda code"

        if tool_name:
            assert result[1][1] == "com.houdinitoolbox__Sop_foo__1_DEFAULT_TOOL"

        else:
            assert result[1][1] == "$HDA_DEFAULT_TOOL"


def _load_test_file(path: pathlib.Path) -> etree._Element:
    """Load a test xml file.

    :param path: The test file path.
    :return: The XML root.

    """
    parser = etree.XMLParser(strip_cdata=False)
    tree = etree.parse(str(path), parser)

    return tree.getroot()



"""This module contains the definition for a python-modernize package runner."""

# =============================================================================
# IMPORTS
# =============================================================================

# Future
from __future__ import annotations

# Standard Library
from typing import TYPE_CHECKING, List

# Houdini Package Runner
import houdini_package_runner.parser
import houdini_package_runner.utils
from houdini_package_runner.discoverers import package
from houdini_package_runner.runners.base import HoudiniPackageRunner

# Imports for type checking.
if TYPE_CHECKING:
    import argparse
    import pathlib

    from houdini_package_runner.items.base import BaseItem


# =============================================================================
# CLASSES
# =============================================================================


class ModernizeRunner(HoudiniPackageRunner):
    """Implementation for a python-modernize package runner."""

    # -------------------------------------------------------------------------
    # PROPERTIES
    # -------------------------------------------------------------------------

    @property
    def name(self) -> str:
        """The runner name used for identification."""
        return "modernize"

    # -------------------------------------------------------------------------
    # METHODS
    # -------------------------------------------------------------------------

    @staticmethod
    def build_parser(parser: argparse.ArgumentParser = None) -> argparse.ArgumentParser:
        """Build a parser for the runner.
        :param parser: Optional parser to add arguments to, otherwise a new one will be created.
        :return: The common parser for the runner.

        """
        if parser is None:
            parser = houdini_package_runner.parser.build_common_parser(
                description="""Run modernize on Houdini package items.

Any unknown args will be passed along to the modernize command.
"""
            )

        return parser

    def init_args_options(self, namespace: argparse.Namespace, extra_args: List[str]):
        """Initialize any extra options from parser data.

        :param namespace: Argument parser namespace.
        :param extra_args: Optional list of extra_args to pass to isort.

        """
        super().init_args_options(namespace, extra_args)

        self._extra_args = extra_args

    def process_path(self, file_path: pathlib.Path, item: BaseItem) -> int:
        """Process a file path.

        :param file_path: The path to format.
        :param item: The item to format.
        :return: The process return code.

        """
        flags = []

        flags.extend(self.extra_args)

        skip_fixers = self.config.get_config_data("skip_fixers", item, file_path)

        command = [
            "python-modernize",
            "--write",
            "--nobackups",
        ]

        if skip_fixers:
            houdini_package_runner.utils.add_or_append_to_flags(
                flags, "--nofix", skip_fixers
            )

        command.extend(flags)

        command.append(str(file_path))

        return houdini_package_runner.utils.execute_subprocess_command(
            command, verbose=self._verbose
        )


# =============================================================================
# FUNCTIONS
# =============================================================================


def main() -> int:
    """Run 'python-modernize' on package files.

    :return: The runner return code.

    """
    parser = ModernizeRunner.build_parser()

    parsed_args, unknown = parser.parse_known_args()

    discoverer = package.init_standard_package_discoverer(parsed_args)

    run_tool = ModernizeRunner(discoverer)
    run_tool.init_args_options(parsed_args, unknown)

    result = run_tool.run()
    return result



"""Test the houdini_package_runner.utils module."""

# =============================================================================
# IMPORTS
# =============================================================================

# Standard Library
import copy
import os

# Third Party
import pytest

# Houdini Package Runner
import houdini_package_runner.utils

# =============================================================================
# TESTS
# =============================================================================


@pytest.mark.parametrize(
    "flags, key, values, sep, expected",
    (
        ([], "--to-skip", ["foo", "bar"], None, ["--to-skip", "foo,bar"]),
        ([], "--to-skip", ["foo", "bar"], "_", ["--to-skip", "foo_bar"]),
        (
            ["--thing"],
            "--to-skip",
            ["foo", "bar"],
            None,
            ["--thing", "--to-skip", "foo,bar"],
        ),
        (
            ["--to-skip", "baz", "--thing"],
            "--to-skip",
            ["foo", "bar"],
            None,
            ["--to-skip", "baz,foo,bar", "--thing"],
        ),
    ),
)
def test_add_or_append_to_flags(flags, key, values, sep, expected):
    """Test houdini_package_runner.utils.add_or_append_to_flags"""
    test_flags = copy.copy(flags)

    if sep is not None:
        houdini_package_runner.utils.add_or_append_to_flags(
            test_flags, key, values, sep
        )

    else:
        houdini_package_runner.utils.add_or_append_to_flags(test_flags, key, values)

    assert test_flags == expected


@pytest.mark.parametrize(
    "verbose, has_pyhome, return_code",
    (
        (False, False, 0),
        (False, False, 1),
        (False, True, 0),
        (True, False, 0),
        (True, False, 1),
    ),
)
def test_execute_subprocess_command(mocker, fp, verbose, has_pyhome, return_code):
    """Test houdini_package_runner.utils.execute_subprocess_command."""
    stdout = "This is stdout" if return_code and not verbose else None
    stderr = "This is stderr" if return_code and not verbose else None

    mock_print = mocker.patch("builtins.print")

    # Make a copy of the env to use for passing to the test.
    dummy_env = os.environ.copy()

    # Remove PYTHONHOME in case it is set, so we can explicitly control whether it is
    # set or not.
    if "PYTHONHOME" in dummy_env:
        del dummy_env["PYTHONHOME"]

    if has_pyhome:
        dummy_env["PYTHONHOME"] = "/some/path"

    mocker.patch("os.environ.copy", return_value=dummy_env)

    cmd = ["hotl", "-t", "foobles", "barbles.otl"]

    fp.register(cmd, returncode=return_code, stdout=stdout, stderr=stderr)

    result = houdini_package_runner.utils.execute_subprocess_command(cmd, verbose)

    assert result == return_code

    # If we are testing with PYTHONHOME in the env, make sure it got removed.
    if has_pyhome:
        assert "PYTHONHOME" not in dummy_env

    if return_code and not verbose:
        mock_print.assert_has_calls(
            [
                mocker.call(stdout),
                mocker.call(stderr),
            ]
        )



"""This module contains runnable items based on Houdini digital assets."""

# =============================================================================
# IMPORTS
# =============================================================================

# Future
from __future__ import annotations

# Standard Library
import json
import os
import pathlib
from typing import TYPE_CHECKING, List, Optional

# Houdini Package Runner
import houdini_package_runner.utils
from houdini_package_runner.items.base import BaseFileItem
from houdini_package_runner.items.dialog_script import DialogScriptItem
from houdini_package_runner.items.filesystem import FileToProcess
from houdini_package_runner.items.xml import ShelfFile

# Imports for type checking.
if TYPE_CHECKING:
    import houdini_package_runner.runners.base


# =============================================================================
# CLASSES
# =============================================================================


class DigitalAssetPythonSection(FileToProcess):
    """Class representing a Python digital asset section to process.

    :param path: The file path to process.
    :param display_name: Display name for test output.
    :param write_back: Whether the item should write itself back to disk.

    """

    def __init__(
        self, path: pathlib.Path, display_name: str, write_back: bool = False
    ) -> None:
        super().__init__(path, write_back=write_back, display_name=display_name)

        section_name = os.path.basename(display_name)

        if section_name not in ("PythonCook", "PythonModule"):
            self.ignored_builtins.append("kwargs")


class ExpandedOperatorType(BaseFileItem):
    """Class representing an operator type inside an expanded digital asset
    folder.

    :param path: The path to the operator specific folder.
    :param name: The name of the operator.
    :param write_back: Whether the item should write itself back to disk.
    :param source_file: Optional source file for the expanded operator definition.

    """

    def __init__(
        self,
        path: pathlib.Path,
        name: str,
        write_back: bool = False,
        source_file: pathlib.PurePath = None,
    ) -> None:
        super().__init__(path, write_back=write_back)

        self._name = name
        self._source_file = source_file

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.name} ({self.path})>"

    # -------------------------------------------------------------------------
    # NON-PUBLIC METHODS
    # -------------------------------------------------------------------------

    def _build_python_section_items(
        self, python_sections: List[pathlib.Path]
    ) -> List[FileToProcess]:
        """Build a list of files to process based on Python sections.

        :param python_sections: The Python section files to process.
        :return: The created Python section items.

        """
        files_to_process = []

        for section_path in python_sections:
            if self._source_file is not None:
                display_name = self._source_file / section_path.name

            else:
                display_name = self.path / section_path.name

            files_to_process.append(
                DigitalAssetPythonSection(
                    section_path,
                    str(display_name),
                    write_back=self.write_back,
                )
            )

        return list(files_to_process)

    def _find_internal_shelf_item(self) -> Optional[ShelfFile]:
        """Find the internal shelf item to process, if possible.

        :return: An item representing the internal tool shelves, if any.

        """

        shelf_path = self.path / "Tools.shelf"

        if not shelf_path.exists():
            return None

        display_name = None

        if self._source_file is not None:
            display_name = str(self._source_file / "Tools.shelf")

        item = ShelfFile(
            shelf_path,
            write_back=self.write_back,
            display_name=display_name,
            tool_name=self.name,
        )

        return item

    def _find_python_section_paths(self) -> List[pathlib.Path]:
        """Build a list of all Python related section files to process.

        This list will include any section files with IsPython in ExtraFileOptions
        and the PythonCook section if it exists.

        :return: A list of Python file sections.

        """
        python_sections = self._get_extra_python_section_files()

        # PythonCook sections are implicitly Python so check for them manually.
        python_cook_path = self.path / "PythonCook"

        if python_cook_path.exists():
            python_sections.append(python_cook_path)

        return python_sections

    def _gather_items(self) -> List[BaseFileItem]:
        """Gather all operator related items.

        This includes any Python sections, shelf tools or Python code from the DialogScript.

        :return: A list of all the processable items associated with the expanded digital asset.

        """
        python_sections = self._find_python_section_paths()

        items_to_process: List[BaseFileItem] = list(
            self._build_python_section_items(python_sections)
        )

        shelf_item = self._find_internal_shelf_item()

        if shelf_item is not None:
            items_to_process.append(shelf_item)

        items_to_process.append(self._get_dialog_script_item())

        return items_to_process

    def _get_dialog_script_item(self) -> DialogScriptItem:
        """Get an item for processing the DialogScript section.

        :return: The DialogScript item.

        """
        dialog_script_path = self.path / "DialogScript"

        if self._source_file is not None:
            display_name = self._source_file.stem.replace("::", "__") + "_DialogScript"

        else:
            display_name = (
                self.name.replace("::", "__").replace("/", "_") + "_DialogScript"
            )

        return DialogScriptItem(
            dialog_script_path, display_name, write_back=self.write_back
        )

    def _get_extra_python_section_files(self) -> List[pathlib.Path]:
        """Find any sections which contain Python code.

        :return: Any section files which contain Python code.

        """
        python_sections = []

        # The extra file options file.
        extra_options = self.path / "ExtraFileOptions"

        if extra_options.exists():
            with extra_options.open() as handle:
                data = json.load(handle)

            for key, values in data.items():
                # Look for sections that are Python.
                if "IsPython" in key:
                    if values["value"]:
                        script_name = key.split("/")[0]
                        section_path = self.path / script_name
                        python_sections.append(section_path)

        return python_sections

    # -------------------------------------------------------------------------
    # PROPERTIES
    # -------------------------------------------------------------------------

    @property
    def name(self) -> str:
        """The name of the operator."""
        return self._name

    # -------------------------------------------------------------------------
    # METHODS
    # -------------------------------------------------------------------------

    def process(
        self, runner: houdini_package_runner.runners.base.HoudiniPackageRunner
    ) -> int:
        """Process the operator items.

        :param runner: The package runner processing the item.
        :return: The process return code.

        """
        items_to_process = self._gather_items()

        result = 0

        for item in items_to_process:
            result |= item.process(runner)
            self.contents_changed |= item.contents_changed

        return result


class DigitalAssetDirectory(BaseFileItem):
    """Class representing an extracted otl file.

    :param path: The path for the item.
    :param write_back: Whether the item should write itself back to disk.
    :param source_file: The binary source file, if any.

    """

    def __init__(
        self, path: pathlib.Path, write_back=False, source_file: pathlib.PurePath = None
    ) -> None:
        super().__init__(path, write_back=write_back)

        self._source_file = source_file

    # -------------------------------------------------------------------------
    # NON-PUBLIC METHODS
    # -------------------------------------------------------------------------

    def _build_operator_list(self) -> List[ExpandedOperatorType]:
        """Build a list of operators in the extracted library.

        :return: A list of operators inside the library.

        """
        operators = []

        operator_dirs = self._find_operator_dirs()

        for definition in operator_dirs:
            directory_name, operator_name = definition
            display_name = None

            if self._source_file is not None:
                display_name = pathlib.PurePath(f"{self._source_file}?{operator_name}")

            operator = ExpandedOperatorType(
                self.path / directory_name,
                operator_name,
                write_back=self.write_back,
                source_file=display_name,
            )

            operators.append(operator)

        return operators

    def _find_operator_dirs(self) -> List[List[str]]:
        """Find a list of operator definition directories based on the Sections.list.

        :return: A list of any directories containing operator definitions.

        """
        sections_list = self.path / "Sections.list"

        if not sections_list.exists():
            raise RuntimeError(
                f"Could not find Sections.list in extracted otl folder: {self.path}"
            )

        with sections_list.open() as handle:
            data = handle.readlines()

        results = []

        for line in data:
            components = line.split()

            if len(components) == 2:
                if (self.path / components[0]).is_dir():
                    results.append(components)

        return results

    # -------------------------------------------------------------------------
    # METHODS
    # -------------------------------------------------------------------------

    def process(
        self, runner: houdini_package_runner.runners.base.HoudiniPackageRunner
    ) -> int:
        """Process the file.

        :param runner: The package runner processing the item.
        :return: The process return code.

        """
        operators = self._build_operator_list()

        result = 0

        for operator in operators:
            result |= operator.process(runner)
            self.contents_changed |= operator.contents_changed

        return result


class BinaryDigitalAssetFile(BaseFileItem):
    """Class representing a binary digital asset file that will be extracted for testing."""

    # -------------------------------------------------------------------------
    # NON-PUBLIC METHODS
    # -------------------------------------------------------------------------

    def _collapse_dir(self, hotl_command: str, target_folder: pathlib.Path) -> None:
        """Collapse the target folder to the digital asset file.

        :param hotl_command: The hotl command to use.
        :param target_folder: The folder to collapse.

        """
        command = [hotl_command, "-l", str(target_folder), str(self.path)]

        return_code = houdini_package_runner.utils.execute_subprocess_command(command)

        if return_code:
            raise RuntimeError(f"An error occurred running the command: {command}")

    def _extract_file(self, hotl_command: str, target_folder: pathlib.Path) -> None:
        """Expand the digital asset file to the target folder.

        :param hotl_command: The hotl command to use.
        :param target_folder: The folder to expand to.

        """
        command = [hotl_command, "-t", str(target_folder), str(self.path)]

        return_code = houdini_package_runner.utils.execute_subprocess_command(command)

        if return_code:
            raise RuntimeError(f"An error occurred running the command: {command}")

    # -------------------------------------------------------------------------
    # METHODS
    # -------------------------------------------------------------------------

    def process(
        self, runner: houdini_package_runner.runners.base.HoudiniPackageRunner
    ) -> int:
        """Process the binary digital asset file.

        :param runner: The package runner processing the item.
        :return: The process return code.

        """
        target_folder = runner.temp_dir / self.path.name

        self._extract_file(runner.hotl_command, target_folder)

        # Create a new DigitalAssetDirectory and process it.
        asset_directory = DigitalAssetDirectory(
            target_folder, write_back=self.write_back, source_file=self.path
        )

        result = asset_directory.process(runner)

        self.contents_changed = asset_directory.contents_changed

        # If writing back, call 'hotl -l' to collapse all the changes.
        if self.write_back and self.contents_changed:
            self._collapse_dir(runner.hotl_command, target_folder)

        return result


# =============================================================================
# FUNCTIONS
# =============================================================================


def is_expanded_digital_asset_dir(path: pathlib.Path) -> bool:
    """Check if a folder is an expanded digital asset.

    We say a folder is an expanded digital asset if it has a "Sections.list" file inside.

    :param path: The folder to check.
    :return: Whether the folder is an expanded digital asset.

    """
    return (path / "Sections.list").is_file()



"""Functioned related to configuration of houdini_package_runner."""

# Future
from __future__ import annotations

# Standard Library
import abc
import importlib.resources
import os
import pathlib
from typing import TYPE_CHECKING, Dict, List, Tuple

# Third Party
import deepmerge
import toml

# Imports for type checking.
if TYPE_CHECKING:
    from houdini_package_runner.items.base import BaseItem


# =============================================================================
# CLASSES
# =============================================================================


class BaseRunnerConfig(abc.ABC):
    """Base class for runner configuration."""

    def __init__(self, runner_name: str):
        self._runner_name = runner_name
        self._data = self.load_config()

    # -------------------------------------------------------------------------
    # PROPERTIES
    # -------------------------------------------------------------------------

    @property
    def data(self) -> dict:
        """The configuration data."""
        return self._data

    @property
    def runner_name(self) -> str:
        """The runner name."""
        return self._runner_name

    # -------------------------------------------------------------------------
    # METHODS
    # -------------------------------------------------------------------------

    @abc.abstractmethod
    def get_config_data(
        self,
        key: str,
        item: BaseItem,
        file_path: pathlib.Path,
    ) -> List:
        """Get config data for an item, and it's file path.

        :param key: The data key.
        :param item: The item to get config data for.
        :param file_path: The item file path to get config data for.
        :return: Any found config data.

        """

    @abc.abstractmethod
    def load_config(self) -> dict:
        """Load the configuration data.

        :return: Any found config data.

        """


class PackageRunnerConfig(BaseRunnerConfig):
    """Config class for default runners.

    This uses the shipped runners.toml file

    """

    # -------------------------------------------------------------------------
    # NON-PUBLIC METHODS
    # -------------------------------------------------------------------------

    def _get_file_config_data(self, file_path: pathlib.Path, key: str) -> List:
        """Get config data based on the file name.

        :param file_path: The file path to get config data for.
        :param key: The data key.
        :return: Any found config data.

        """
        data = []

        file_config = self.data.get("file", {})

        for file_name, file_options in file_config.items():
            if file_name == file_path.name:
                data.extend(file_options.get(key, []))

        return data

    def _get_item_config_data(self, item: BaseItem, key: str) -> List:
        """Get config data based on an item.

        :param item: The item to get config data for.
        :param key: The data key.
        :return: Any found config data.

        """
        config_names = build_config_item_list(item)

        data = []

        item_config = self.data.get("item", {})

        for config_name in config_names:
            config_items = item_config.get(config_name, {})
            data.extend(config_items.get(key, []))

        if item.is_test_item:
            test_items = item_config.get("test_item", {})
            data.extend(test_items.get(key, []))

        return data

    # -------------------------------------------------------------------------
    # METHODS
    # -------------------------------------------------------------------------

    def get_config_data(
        self,
        key: str,
        item: BaseItem,
        file_path: pathlib.Path,
    ) -> List:
        """Get config data for an item, and it's file path.

        :param key: The data key.
        :param item: The item to get config data for.
        :param file_path: The item file path to get config data for.
        :return: Any found config data.

        """
        data = []

        data.extend(self._get_item_config_data(item, key))
        data.extend(self._get_file_config_data(file_path, key))

        return data

    def load_config(self) -> dict:
        """Load the configuration data.

        :return: Any found config data.

        """
        return _load_default_runner_config(self.runner_name)


# =============================================================================
# NON-PUBLIC FUNCTIONS
# =============================================================================


def _find_config_files() -> List[pathlib.Path]:
    """Find any config files to read.

    This uses HOUDINI_PACKAGE_RUNNER_CONFIG_PATH to find any specified config files,
    otherwise it will return the packaged resource.

    :return: One or more config files.

    """
    path_env = os.getenv("HOUDINI_PACKAGE_RUNNER_CONFIG_PATH")

    # Load the packaged config file if none is specified.
    if path_env is None:
        with importlib.resources.path("houdini_package_runner", "runners.toml") as path:
            paths = [path]

    else:
        path_components = path_env.split(os.path.pathsep)

        paths = [
            pathlib.Path(component)
            for component in path_components
            if os.path.exists(component)
        ]

    return paths


def _get_base_classes(cls: type) -> List[type]:
    """Get a list of all the base classes of a class.

    This will exclude the base "object" class.

    :param cls: The class to get the base classes for.
    :return: A list of all the base classes.

    """
    bases = [base for base in cls.__bases__ if not base == object]

    for base in bases:
        bases.extend(_get_base_classes(base))

    return bases


def _load_default_runner_config(runner_name: str) -> dict:
    """Load the configuration for a runner.

    :param runner_name: The name of the runner.
    :return: The configuration dictionary for the runner.

    """
    paths = _find_config_files()

    data: Dict = {}

    # In the event of multiple files, process them all and do a deep merge, preferring
    # data already found over new values in the event of collision.
    for path in paths:
        with path.open("r", encoding="utf-8") as handle:
            path_data = toml.load(handle)
            deepmerge.conservative_merger.merge(data, path_data)

    return data.get(runner_name, {})


# =============================================================================
# FUNCTIONS
# =============================================================================


def build_config_item_list(item: BaseItem) -> Tuple[str, ...]:
    """Build a list of the item's class and super class names.

    :param item: The item to get class names for.
    :return: A list containing the item's class name as well as any super class names.

    """
    bases = [base.__name__ for base in _get_base_classes(type(item))]

    names = [type(item).__name__] + bases

    return tuple(names)



"""This module contains the definition for an flake8 package runner."""

# =============================================================================
# IMPORTS
# =============================================================================

# Future
from __future__ import annotations

# Standard Library
import pathlib
from typing import TYPE_CHECKING, List

# Houdini Package Runner
import houdini_package_runner.parser
import houdini_package_runner.utils
from houdini_package_runner.discoverers import package
from houdini_package_runner.runners.base import HoudiniPackageRunner

# Imports for type checking.
if TYPE_CHECKING:
    import argparse

    from houdini_package_runner.config import BaseRunnerConfig
    from houdini_package_runner.discoverers.base import BaseItemDiscoverer
    from houdini_package_runner.items.base import BaseItem


# =============================================================================
# CLASSES
# =============================================================================


class Flake8Runner(HoudiniPackageRunner):
    """Implementation for a flake8 package runner.

    :param discoverer: The item discoverer used by the runner.
    :param runner_config: Optional BaseRunnerConfig object.

    """

    def __init__(
        self, discoverer: BaseItemDiscoverer, runner_config: BaseRunnerConfig = None
    ) -> None:
        super().__init__(discoverer, write_back=True, runner_config=runner_config)

        self._ignored: List[str] = []

    # -------------------------------------------------------------------------
    # PROPERTIES
    # -------------------------------------------------------------------------

    @property
    def name(self) -> str:
        """The runner name used for identification."""
        return "flake8"

    # -------------------------------------------------------------------------
    # METHODS
    # -------------------------------------------------------------------------

    @staticmethod
    def build_parser(parser: argparse.ArgumentParser = None) -> argparse.ArgumentParser:
        """Build a parser for the runner.

        :param parser: Optional parser to add arguments to, otherwise a new one will be created.
        :return: The common parser for the runner.

        """
        if parser is None:
            parser = houdini_package_runner.parser.build_common_parser(
                description="""Run flake8 on Houdini package items.

Any unknown args will be passed along to the flake8 command.
"""
            )

        parser.add_argument(
            "--config",
            action="store",
            help="Specify a configuration file",
        )

        parser.add_argument("--ignore", action="store", help="Tests to ignore.")

        return parser

    def init_args_options(self, namespace: argparse.Namespace, extra_args: List[str]):
        """Initialize any extra options from parser data.

        :param namespace: Argument parser namespace.
        :param extra_args: Optional list of extra_args to pass to isort.

        """
        super().init_args_options(namespace, extra_args)

        if namespace.config:
            extra_args.insert(0, f"--config={namespace.config}")

        if namespace.ignore:
            self._ignored = namespace.ignore.split(",")

        self._extra_args = extra_args

    def process_path(self, file_path: pathlib.Path, item: BaseItem) -> int:
        """Process a file path.

        :param file_path: The path to format.
        :param item: The item to format.
        :return: The process return code.

        """
        command = [
            "flake8",
        ]

        to_ignore = []

        if self._ignored:
            to_ignore.extend(self._ignored)

        known_builtins: List[str] = item.ignored_builtins

        to_ignore.extend(self.config.get_config_data("to_ignore", item, file_path))

        command.extend(self.config.get_config_data("command", item, file_path))

        known_builtins.extend(
            self.config.get_config_data("known_builtins", item, file_path)
        )

        if known_builtins:
            houdini_package_runner.utils.add_or_append_to_flags(
                command, "--builtins", known_builtins
            )

        if to_ignore:
            command.append(f"--ignore={','.join(to_ignore)}")

        command.extend(self.extra_args)

        command.append(str(file_path))

        return houdini_package_runner.utils.execute_subprocess_command(
            command, verbose=self._verbose
        )


# =============================================================================
# FUNCTIONS
# =============================================================================


def main() -> int:
    """Run 'flake8' on package files.

    :return: The runner return code.

    """
    parser = Flake8Runner.build_parser()

    parsed_args, unknown = parser.parse_known_args()

    discoverer = package.init_standard_package_discoverer(parsed_args)

    run_tool = Flake8Runner(discoverer)
    run_tool.init_args_options(parsed_args, unknown)

    result = run_tool.run()
    return result



"""Test the houdini_package_runner.runners.flake8.runner module."""

# =============================================================================
# IMPORTS
# =============================================================================

# Standard Library
import argparse
import copy
import pathlib

# Third Party
import pytest

# Houdini Package Runner
import houdini_package_runner.config
import houdini_package_runner.items.base
import houdini_package_runner.items.dialog_script
import houdini_package_runner.items.digital_asset
import houdini_package_runner.items.filesystem
import houdini_package_runner.items.xml
import houdini_package_runner.runners.base
import houdini_package_runner.runners.flake8.runner
from houdini_package_runner.discoverers.base import BaseItemDiscoverer

# =============================================================================
# FIXTURES
# =============================================================================


@pytest.fixture
def init_runner(mocker):
    """Initialize a dummy Flake8Runner for testing."""
    mocker.patch.multiple(
        houdini_package_runner.runners.flake8.runner.Flake8Runner,
        __init__=lambda x, y: None,
    )

    def _create():
        return houdini_package_runner.runners.flake8.runner.Flake8Runner(None)

    return _create


# =============================================================================
# TESTS
# =============================================================================


class TestFlake8Runner:
    """Test houdini_package_runner.runners.flake8.runner.Flake8Runner."""

    @pytest.mark.parametrize("pass_optional", (False, True))
    def test___init__(self, mocker, pass_optional):
        """Test object initialization."""
        mock_discoverer = mocker.MagicMock(spec=BaseItemDiscoverer)

        mock_super_init = mocker.patch.object(
            houdini_package_runner.runners.flake8.runner.HoudiniPackageRunner,
            "__init__",
        )

        mock_config = (
            mocker.MagicMock(spec=houdini_package_runner.config.PackageRunnerConfig)
            if pass_optional
            else None
        )

        if pass_optional:
            inst = houdini_package_runner.runners.flake8.runner.Flake8Runner(
                mock_discoverer, runner_config=mock_config
            )

        else:
            inst = houdini_package_runner.runners.flake8.runner.Flake8Runner(
                mock_discoverer
            )

        assert inst._ignored == []

        mock_super_init.assert_called_with(
            mock_discoverer, write_back=True, runner_config=mock_config
        )

    # Properties

    def test_name(self, init_runner):
        """Test Flake8Runner.name."""
        inst = init_runner()

        assert inst.name == "flake8"

        with pytest.raises(AttributeError):
            inst.name = []

    # Methods

    @pytest.mark.parametrize("pass_parser", (True, False))
    def test_build_parser(self, mocker, pass_parser):
        """Test Flake8Runner.build_parser."""
        mock_parser = mocker.MagicMock(spec=argparse.ArgumentParser)

        mock_build = mocker.patch(
            "houdini_package_runner.parser.build_common_parser",
            return_value=mock_parser,
        )

        if pass_parser:
            result = (
                houdini_package_runner.runners.flake8.runner.Flake8Runner.build_parser(
                    parser=mock_parser
                )
            )

            mock_build.assert_not_called()

        else:
            result = (
                houdini_package_runner.runners.flake8.runner.Flake8Runner.build_parser()
            )
            mock_build.assert_called()

        assert result == mock_parser

        result.add_argument.assert_has_calls(
            [
                mocker.call(
                    "--config",
                    action="store",
                    help="Specify a configuration file",
                ),
                mocker.call("--ignore", action="store", help="Tests to ignore."),
            ]
        )

    @pytest.mark.parametrize("has_options_set", (True, False))
    def test_init_args_options(self, mocker, init_runner, has_options_set):
        """Test Flake8Runner.init_args_options."""
        mock_namespace = mocker.MagicMock()

        if has_options_set:
            mock_namespace.config = "config_path"
            mock_namespace.ignore = "foo,bar"

        else:
            mock_namespace.config = None
            mock_namespace.ignore = None

        extra_args = ["arg1", "arg2"]

        expected_extra_args = copy.copy(extra_args)

        if has_options_set:
            expected_extra_args.insert(0, f"--config={mock_namespace.config}")

        mock_super_init = mocker.patch.object(
            houdini_package_runner.runners.flake8.runner.HoudiniPackageRunner,
            "init_args_options",
        )

        inst = init_runner()
        inst._ignored = []

        inst.init_args_options(mock_namespace, extra_args)

        mock_super_init.assert_called_with(mock_namespace, extra_args)

        assert inst._extra_args == extra_args

        if has_options_set:
            assert inst._ignored == ["foo", "bar"]

        else:
            assert inst._ignored == []

    @pytest.mark.parametrize(
        "has_ignored, has_builtins",
        (
            (True, True),
            (False, False),
            (False, False),
        ),
    )
    def test_process_path(self, mocker, init_runner, has_ignored, has_builtins):
        """Test Flake8Runner.process_path."""
        mock_path = mocker.MagicMock(spec=pathlib.Path)

        mock_item = mocker.MagicMock(spec=houdini_package_runner.items.base.BaseItem)

        mock_item.ignored_builtins = ["hou"] if has_builtins else []

        to_ignore = ["ABC"] if has_builtins else []
        extra_command = ["--arg1", "arg1_val"]
        builtins = ["bob"] if has_builtins else []

        mock_config = mocker.MagicMock(
            spec=houdini_package_runner.config.PackageRunnerConfig
        )
        mock_config.get_config_data.side_effect = [to_ignore, extra_command, builtins]

        mocker.patch.object(
            houdini_package_runner.runners.flake8.runner.Flake8Runner,
            "config",
            mock_config,
        )

        mock_add_flags = mocker.patch(
            "houdini_package_runner.utils.add_or_append_to_flags"
        )

        mock_execute = mocker.patch(
            "houdini_package_runner.utils.execute_subprocess_command"
        )

        extra_args = ["--flag", "arg"]

        mocker.patch.object(
            houdini_package_runner.runners.flake8.runner.Flake8Runner,
            "extra_args",
            extra_args,
        )

        mock_verbose = mocker.MagicMock(spec=bool)

        expected_ignored = []

        inst = init_runner()
        inst._ignored = []
        inst._verbose = mock_verbose

        if has_ignored:
            inst._ignored = ["A123"]
            expected_ignored = inst._ignored + to_ignore

        inst.process_path(mock_path, mock_item)

        expected_args = ["flake8"] + extra_command

        if has_builtins:
            # Only do assert_called() here as the command list will change and be inaccurate.
            mock_add_flags.assert_called()

        if expected_ignored:
            expected_args.append(f"--ignore={','.join(expected_ignored)}")

        expected_args.extend(extra_args)
        expected_args.append(str(mock_path))

        mock_config.get_config_data.assert_has_calls(
            [
                mocker.call("to_ignore", mock_item, mock_path),
                mocker.call("command", mock_item, mock_path),
                mocker.call("known_builtins", mock_item, mock_path),
            ]
        )

        mock_execute.assert_called_with(expected_args, verbose=mock_verbose)


def test_main(mocker):
    """Test houdini_package_runner.runners.flake8.runner.main."""
    mock_parsed = mocker.MagicMock(spec=argparse.Namespace)
    mock_unknown = mocker.MagicMock(spec=list)

    mock_parser = mocker.MagicMock(spec=argparse.ArgumentParser)
    mock_parser.parse_known_args.return_value = (mock_parsed, mock_unknown)

    mock_discoverer = mocker.MagicMock(spec=BaseItemDiscoverer)
    mock_init = mocker.patch(
        "houdini_package_runner.runners.flake8.runner.package.init_standard_package_discoverer",
        return_value=mock_discoverer,
    )

    mock_runner = mocker.MagicMock(
        spec=houdini_package_runner.runners.flake8.runner.Flake8Runner
    )

    mock_runner_init = mocker.patch(
        "houdini_package_runner.runners.flake8.runner.Flake8Runner",
        return_value=mock_runner,
    )
    mock_runner_init.build_parser.return_value = mock_parser

    result = houdini_package_runner.runners.flake8.runner.main()
    assert result == mock_runner.run.return_value

    mock_init.assert_called_with(mock_parsed)

    mock_runner_init.assert_called_with(mock_discoverer)
    mock_runner.init_args_options.assert_called_with(mock_parsed, mock_unknown)
    mock_runner.run.assert_called()



"""This module contains the definition for an isort package runner."""

# =============================================================================
# IMPORTS
# =============================================================================

# Future
from __future__ import annotations

# Standard Library
import importlib.resources
import os
import pathlib
from configparser import ConfigParser
from typing import TYPE_CHECKING, List, Optional

# Houdini Package Runner
import houdini_package_runner.parser
import houdini_package_runner.utils
from houdini_package_runner.discoverers import package
from houdini_package_runner.runners.base import HoudiniPackageRunner

# Imports for type checking.
if TYPE_CHECKING:
    import argparse

    from houdini_package_runner.config import BaseRunnerConfig
    from houdini_package_runner.discoverers.base import BaseItemDiscoverer
    from houdini_package_runner.items.base import BaseItem


# =============================================================================
# CLASSES
# =============================================================================


class IsortRunner(HoudiniPackageRunner):
    """Implementation for an isort package runner.

    :param discoverer: The item discoverer used by the runner.
    :param runner_config: Optional BaseRunnerConfig object.

    """

    def __init__(
        self, discoverer: BaseItemDiscoverer, runner_config: BaseRunnerConfig = None
    ) -> None:
        super().__init__(discoverer, write_back=True, runner_config=runner_config)

        self._config_file: Optional[str] = None

    def _generate_config(self, namespace: argparse.Namespace):
        """Generate a .isort config file for the operation.

        :param namespace: The command argparse namespace.
        :return: The path to the generated config file.

        """
        config = _load_template_config()

        self._process_config(config, namespace)

        return _save_template_config(config, self.temp_dir)

    def _get_first_party_header(  # pylint: disable=no-self-use
        self, package_names: str
    ) -> str:
        """Get a header for the First Party section.

        This will take the first module name in the comma separated list and use title() on it.

        :param package_names: A comma separated list of first party package names.
        :return: A header name.

        """
        return package_names.split(",")[0].replace("_", " ").title()

    def _get_first_party_packages(self, namespace: argparse.Namespace) -> Optional[str]:
        """Find a list of known first party module names, if any.

        :param namespace: The command argparse namespace.
        :return: A comma separated list of first party names, otherwise None.

        """
        first_party_packages = None

        if namespace.package_names is not None:
            first_party_packages = namespace.package_names

        # If no first party package names were passed we can try and figure out any
        # based on the Python root.
        elif namespace.python_root is not None:
            python_root = self.discoverer.path / namespace.python_root

            if python_root.exists():
                first_party_packages = _find_python_packages_from_path(python_root)

        return first_party_packages

    def _get_houdini_names(  # pylint: disable=no-self-use
        self, namespace: argparse.Namespace
    ) -> str:
        """Find a list of known Houdini shipped module names.

        :param namespace: The command argparse namespace.
        :return: A comma separated list of Houdini Python modules.

        """
        hfs_path = pathlib.Path(os.path.expandvars(namespace.hfs_path))

        return ",".join(_find_known_houdini(hfs_path))

    def _process_config(
        self,
        config: ConfigParser,
        namespace: argparse.Namespace,
    ):
        """Process the ConfigParser object.

        :param config: The configuration object.
        :param namespace: The command argparse namespace.
        :return:

        """
        settings = config["settings"]

        settings["known_houdini"] = self._get_houdini_names(namespace)

        first_party_packages = self._get_first_party_packages(namespace)

        if first_party_packages is not None:
            settings["known_first_party"] = first_party_packages

            settings["import_heading_firstparty"] = self._get_first_party_header(
                first_party_packages
            )

    # -------------------------------------------------------------------------
    # PROPERTIES
    # -------------------------------------------------------------------------

    @property
    def config_file(self) -> Optional[str]:
        """Optional isort config file."""
        return self._config_file

    @config_file.setter
    def config_file(self, setting_file: str):
        self._config_file = setting_file

    # -------------------------------------------------------------------------

    @property
    def name(self) -> str:
        """The runner name used for identification."""
        return "isort"

    # -------------------------------------------------------------------------
    # METHODS
    # -------------------------------------------------------------------------

    @staticmethod
    def build_parser(parser: argparse.ArgumentParser = None) -> argparse.ArgumentParser:
        """Build a parser for the runner.

        :param parser: Optional parser to add arguments to, otherwise a new one will be created.
        :return: The common parser for the runner.

        """
        if parser is None:
            parser = houdini_package_runner.parser.build_common_parser(
                description="""Run isort on Houdini package items.

Any unknown args will be passed along to the isort command.
"""
            )

        parser.add_argument(
            "--config-file",
            action="store",
            default=".isort.cfg",
            help="Optional config file to pass to isort commands",
        )

        parser.add_argument(
            "--hfs-path",
            action="store",
            default="$HFS",
            help="Path to a Houdini install directory for known Houdini modules",
        )

        parser.add_argument(
            "--package-names", action="store", help="Known first party package names"
        )

        return parser

    def init_args_options(self, namespace: argparse.Namespace, extra_args: List[str]):
        """Initialize any extra options from parser data.

        :param namespace: Argument parser namespace.
        :param extra_args: Optional list of extra_args to pass to isort.

        """
        super().init_args_options(namespace, extra_args)

        config_file = namespace.config_file

        if config_file is not None:
            if (self.discoverer.path / config_file).exists():
                self.config_file = self.discoverer.path / config_file

        self._extra_args = extra_args

        if self.config_file is None:
            self.config_file = self._generate_config(namespace)

    def process_path(self, file_path: pathlib.Path, item: BaseItem) -> int:
        """Process a file path.

        :param file_path: The path to format.
        :param item: The item to format.
        :return: The process return code.

        """
        command = [
            "isort",
        ]

        if self.config_file is not None:
            command.extend(("--sp", str(self.config_file)))

        command.extend(self.extra_args)

        command.append(str(file_path))

        return houdini_package_runner.utils.execute_subprocess_command(
            command, verbose=self._verbose
        )


# =============================================================================
# NON-PUBLIC FUNCTIONS
# =============================================================================


def _find_known_houdini(hfs_path: pathlib.Path) -> List[str]:
    """Find a list of known Houdini Python modules.

    :param hfs_path: The path to a Houdini install directory ($HFS)
    :return: A list of Houdini module names.

    """
    module_names = []

    python_libs_paths = (hfs_path / "houdini").glob("python*libs")

    for path in python_libs_paths:
        module_names.extend(_find_python_modules(path))

    soho_paths = (hfs_path / "houdini" / "soho").glob("python*")

    for path in soho_paths:
        module_names.extend(_find_python_modules(path))

    return sorted(set(module_names))


def _find_python_modules(folder: pathlib.Path) -> List[str]:
    """Find modules in a directory.

    :param folder: The directory to look for Python modules in.
    :return: A list of Python module names.

    """
    module_names = []

    for child in folder.iterdir():
        if child.is_file() and child.suffix in (".py", ".so"):
            module_names.append(child.stem)

        elif child.is_dir() and not child.stem.startswith("__"):
            module_names.append(child.stem)

    return sorted(module_names)


def _find_python_packages_from_path(search_root: pathlib.Path) -> Optional[str]:
    """Attempt to find any Python package names under the search path.

    :param search_root: The path to search under.
    :return: Any found package names.

    """
    package_names = []

    # Check for any folders under the root.
    for path in search_root.iterdir():
        if path.is_dir() and (path / "__init__.py").exists():
            package_names.append(path.stem)

    return ",".join(sorted(package_names)) if package_names else None


def _load_template_config() -> ConfigParser:
    """Load the template config file.

    :return: The loaded default config file.

    """
    config = ConfigParser()

    with importlib.resources.open_text(
        "houdini_package_runner.runners.isort", "isort.cfg"
    ) as contents:
        config.read_file(contents)

    return config


def _save_template_config(config: ConfigParser, temp_dir: pathlib.Path):
    """Save the config to a temporary file for running.

    :param config: The configuration to save.
    :param temp_dir: The runner temporary directory.
    :return: The saved configuration file path.

    """
    file_path = temp_dir / ".isort.cfg"

    with open(file_path, "w", encoding="utf-8") as handle:
        config.write(handle)

    return file_path


# =============================================================================
# FUNCTIONS
# =============================================================================


def main() -> int:
    """Run 'isort' on package files.

    :return: The runner return code.

    """
    parser = IsortRunner.build_parser()

    parsed_args, unknown = parser.parse_known_args()

    discoverer = package.init_standard_package_discoverer(parsed_args)

    run_tool = IsortRunner(discoverer)
    run_tool.init_args_options(parsed_args, unknown)

    result = run_tool.run()
    return result



"""Test the houdini_package_runner.discoverers.package module."""

# =============================================================================
# IMPORTS
# =============================================================================

# Standard Library
import argparse
import pathlib

# Third Party
import pytest

# Houdini Package Runner
import houdini_package_runner.discoverers.base
import houdini_package_runner.discoverers.package
import houdini_package_runner.items.digital_asset
import houdini_package_runner.items.filesystem
import houdini_package_runner.items.xml

# =============================================================================
# TESTS
# =============================================================================


class TestPackageItemDiscoverer:
    """Test houdini_package_runner.discoverers.package.PackageItemDiscoverer."""

    # Object construction

    @pytest.mark.parametrize("has_items", (False, True))
    def test___init__(self, mocker, has_items):
        """Test object initialization."""
        mock_path = mocker.MagicMock(spec=pathlib.Path)

        mock_file_path = mocker.MagicMock(spec=pathlib.Path)
        mock_file_path.is_file.return_value = True
        mock_file_path.is_dir.return_value = False

        mock_file1 = mocker.MagicMock(
            spec=houdini_package_runner.items.filesystem.FileToProcess
        )
        mock_dir = mocker.MagicMock(
            spec=houdini_package_runner.items.filesystem.DirectoryToProcess
        )

        mock_process_dir = mocker.patch(
            "houdini_package_runner.discoverers.package.process_directory"
        )
        mock_process_dir.side_effect = (mock_dir, None)

        mock_houdini_item = mocker.MagicMock(
            spec=houdini_package_runner.items.filesystem.HoudiniDirectoryItem
        )

        mock_get_houdini = mocker.patch(
            "houdini_package_runner.discoverers.package.get_houdini_items"
        )
        mock_get_houdini.return_value = [mock_houdini_item] if has_items else []

        mock_file_to_process = mocker.patch(
            "houdini_package_runner.items.filesystem.FileToProcess"
        )

        if has_items:
            items = [mock_file1]

            houdini_items = ["scripts"]

            mock_dir1 = mocker.MagicMock(spec=pathlib.Path)
            mock_dir1.is_file.return_value = False
            mock_dir1.is_dir.return_value = True

            mock_dir2 = mocker.MagicMock(spec=pathlib.Path)
            mock_dir2.is_file.return_value = False
            mock_dir2.is_dir.return_value = True

            extra_paths = [mock_file_path, mock_dir1, mock_dir2]

            inst = houdini_package_runner.discoverers.package.PackageItemDiscoverer(
                mock_path,
                houdini_items,
                extra_paths=extra_paths,
                items=items,
            )

            assert inst.items == [
                mock_file1,
                mock_houdini_item,
                mock_file_to_process.return_value,
                mock_dir,
            ]

            mock_file_to_process.assert_called_with(mock_file_path)
            mock_get_houdini.assert_called_with(houdini_items, inst.path)

        else:
            inst = houdini_package_runner.discoverers.package.PackageItemDiscoverer(
                mock_path,
                houdini_items=[],
            )

            assert inst.items == []


def test_get_digital_asset_items(shared_datadir):
    """Test houdini_package_runner.discoverers.package.get_digital_asset_items."""
    test_path = shared_datadir / "get_digital_asset_items"

    results = houdini_package_runner.discoverers.package.get_digital_asset_items(
        test_path
    )

    assert len(results) == 3

    expanded_dir_path = test_path / "expanded_dir"
    nodetype_otl_path = test_path / "nodetype.otl"
    operator_hda_path = test_path / "operator.hda"

    for item in results:
        if item.path in (nodetype_otl_path, operator_hda_path):
            assert isinstance(
                item, houdini_package_runner.items.digital_asset.BinaryDigitalAssetFile
            )

        elif item.path == expanded_dir_path:
            assert isinstance(
                item, houdini_package_runner.items.digital_asset.DigitalAssetDirectory
            )


def test_get_houdini_items(mocker, shared_datadir):
    """Test houdini_package_runner.discoverers.package.get_houdini_items."""
    mock_asset_item = mocker.MagicMock(
        spec=houdini_package_runner.items.digital_asset.BinaryDigitalAssetFile
    )
    mock_get_asset_items = mocker.patch(
        "houdini_package_runner.discoverers.package.get_digital_asset_items",
        return_value=[mock_asset_item],
    )

    mock_tool_item = mocker.MagicMock(spec=houdini_package_runner.items.xml.ShelfFile)
    mock_get_tool_items = mocker.patch(
        "houdini_package_runner.discoverers.package.get_tool_items",
        return_value=[mock_tool_item],
    )

    mock_panel_item = mocker.MagicMock(
        spec=houdini_package_runner.items.xml.PythonPanelFile
    )
    mock_get_panel_items = mocker.patch(
        "houdini_package_runner.discoverers.package.get_python_panel_items",
        return_value=[mock_panel_item],
    )

    mock_menu_item = mocker.MagicMock(spec=houdini_package_runner.items.xml.MenuFile)
    mock_get_menu_items = mocker.patch(
        "houdini_package_runner.discoverers.package.get_menu_items",
        return_value=[mock_menu_item],
    )

    mock_pydir_item = mocker.patch(
        "houdini_package_runner.items.filesystem.HoudiniDirectoryItem"
    )

    mock_dir_item = mocker.MagicMock(
        spec=houdini_package_runner.items.filesystem.DirectoryToProcess
    )

    mock_process = mocker.patch(
        "houdini_package_runner.discoverers.package.process_directory",
        side_effect=(mock_dir_item, None),
    )

    test_path = shared_datadir / "get_houdini_items"

    item_names = [
        "",
        "otls",
        "hda",
        "directory_item",
        "empty_directory_item",
        "pythonXlibs",
        "toolbar",
        "python_panels",
        "menus",
        "some_file",
    ]

    items = houdini_package_runner.discoverers.package.get_houdini_items(
        item_names, test_path
    )

    expected = [
        mock_asset_item,
        mock_dir_item,
        mock_pydir_item.return_value,
        mock_tool_item,
        mock_panel_item,
        mock_menu_item,
    ]

    assert items == expected

    mock_get_asset_items.assert_called_with(test_path / "otls")
    mock_get_tool_items.assert_called_with(test_path / "toolbar")
    mock_get_panel_items.assert_called_with(test_path / "python_panels")
    mock_get_menu_items.assert_called_with(test_path)

    mock_pydir_item.assert_called_with(
        test_path / "python3.7libs", traverse_children=True
    )

    mock_process.assert_has_calls(
        [
            mocker.call(test_path / "directory_item"),
            mocker.call(test_path / "empty_directory_item"),
        ]
    )


def test_get_menu_items(mocker):
    """Test houdini_package_runner.discoverers.package.get_menu_items."""
    mock_menu_file = mocker.patch("houdini_package_runner.items.xml.MenuFile")

    mock_menu_path = mocker.MagicMock(spec=pathlib.Path)

    mock_houdini_root = mocker.MagicMock(spec=pathlib.Path)
    mock_houdini_root.glob.return_value = [mock_menu_path]

    result = houdini_package_runner.discoverers.package.get_menu_items(
        mock_houdini_root
    )

    assert result == [mock_menu_file.return_value]

    mock_houdini_root.glob.assert_called_with("*.xml")
    mock_menu_file.assert_called_with(mock_menu_path)


def test_get_python_panel_items(mocker):
    """Test houdini_package_runner.discoverers.package.get_python_panel_items."""
    mock_panel_file = mocker.patch("houdini_package_runner.items.xml.PythonPanelFile")

    mock_panel_path = mocker.MagicMock(spec=pathlib.Path)

    mock_panel_root = mocker.MagicMock(spec=pathlib.Path)
    mock_panel_root.glob.return_value = [mock_panel_path]

    result = houdini_package_runner.discoverers.package.get_python_panel_items(
        mock_panel_root
    )

    assert result == [mock_panel_file.return_value]

    mock_panel_root.glob.assert_called_with("*.pypanel")
    mock_panel_file.assert_called_with(mock_panel_path)


def test_get_tool_items(mocker):
    """Test houdini_package_runner.discoverers.package.get_tool_items."""
    mock_shelf_file = mocker.patch("houdini_package_runner.items.xml.ShelfFile")

    mock_shelf_path = mocker.MagicMock(spec=pathlib.Path)

    mock_toolbar_path = mocker.MagicMock(spec=pathlib.Path)
    mock_toolbar_path.glob.return_value = [mock_shelf_path]

    result = houdini_package_runner.discoverers.package.get_tool_items(
        mock_toolbar_path
    )

    assert result == [mock_shelf_file.return_value]

    mock_toolbar_path.glob.assert_called_with("*.shelf")
    mock_shelf_file.assert_called_with(mock_shelf_path)


def test_init_standard_package_discoverer(
    mocker,
):
    """Test houdini_package_runner.discoverers.package.init_standard_package_discoverer."""
    mock_discoverer = mocker.patch(
        "houdini_package_runner.discoverers.package.PackageItemDiscoverer"
    )

    mock_root = mocker.MagicMock(spec=pathlib.Path)
    mock_houdini_root = mocker.MagicMock(spec=pathlib.Path)
    mock_extra_paths = mocker.MagicMock(spec=list)
    mock_houdini_items = mocker.MagicMock(spec=list)

    mock_parse = mocker.patch(
        "houdini_package_runner.parser.process_common_arg_settings"
    )
    mock_parse.return_value = (
        mock_root,
        mock_houdini_root,
        mock_extra_paths,
        mock_houdini_items,
    )

    mock_namespace = mocker.MagicMock(spec=argparse.Namespace)

    result = (
        houdini_package_runner.discoverers.package.init_standard_package_discoverer(
            mock_namespace
        )
    )

    assert result == mock_discoverer.return_value

    mock_parse.assert_called_with(mock_namespace)

    mock_discoverer.assert_called_with(
        mock_houdini_root,
        houdini_items=mock_houdini_items,
        extra_paths=mock_extra_paths,
    )


@pytest.mark.parametrize(
    "test_path, expected",
    (
        (
            "package_dir",
            houdini_package_runner.items.filesystem.PythonPackageDirectoryItem,
        ),
        ("python", houdini_package_runner.items.filesystem.PythonPackageDirectoryItem),
        (
            "scripts",
            houdini_package_runner.items.filesystem.HoudiniScriptsDirectoryItem,
        ),
        ("tests", houdini_package_runner.items.filesystem.DirectoryToProcess),
        ("other", houdini_package_runner.items.filesystem.DirectoryToProcess),
    ),
)
def test_process_directory(shared_datadir, test_path, expected):
    """Test houdini_package_runner.discoverers.package.process_directory."""
    test_dir = shared_datadir / "process_directory" / test_path

    result = houdini_package_runner.discoverers.package.process_directory(test_dir)

    assert isinstance(result, expected)

    # Items which aren't Python packages should have 'traverse_children' set.
    if not isinstance(
        result, houdini_package_runner.items.filesystem.PythonPackageDirectoryItem
    ):
        assert result.traverse_children

    if test_path == "tests":
        assert result.is_test_item



"""Test the houdini_package_runner.runners.base module."""

# =============================================================================
# IMPORTS
# =============================================================================

# Standard Library
import argparse
import pathlib

# Third Party
import pytest

# Houdini Package Runner
import houdini_package_runner.config
import houdini_package_runner.items.filesystem
import houdini_package_runner.runners.base
from houdini_package_runner.discoverers.base import BaseItemDiscoverer

# =============================================================================
# FIXTURES
# =============================================================================


@pytest.fixture
def init_runner(mocker):
    """Initialize a dummy HoudiniPackageRunner for testing."""
    mocker.patch.multiple(
        houdini_package_runner.runners.base.HoudiniPackageRunner,
        __abstractmethods__=set(),
        __init__=lambda x, y, z: None,
    )

    def _create():
        return houdini_package_runner.runners.base.HoudiniPackageRunner(None, None)

    return _create


# =============================================================================
# TESTS
# =============================================================================


class TestHoudiniPackageRunner:
    """Test houdini_package_runner.runners.base.HoudiniPackageRunner."""

    # Object construction

    @pytest.mark.parametrize("pass_optional", (False, True))
    def test___init__(self, mocker, remove_abstract_methods, pass_optional):
        """Test object initialization."""
        remove_abstract_methods(
            houdini_package_runner.runners.base.HoudiniPackageRunner
        )

        mock_write_back = mocker.MagicMock(spec=bool) if pass_optional else False
        mock_discoverer = mocker.MagicMock(spec=BaseItemDiscoverer)
        mock_mkdtemp = mocker.patch("tempfile.mkdtemp", return_value="/path/to/temp")

        mock_config = (
            mocker.MagicMock(spec=houdini_package_runner.config.PackageRunnerConfig)
            if pass_optional
            else None
        )

        mock_init_config = mocker.patch(
            "houdini_package_runner.runners.base.PackageRunnerConfig"
        )

        if pass_optional:
            inst = houdini_package_runner.runners.base.HoudiniPackageRunner(
                mock_discoverer, write_back=mock_write_back, runner_config=mock_config
            )

        else:
            inst = houdini_package_runner.runners.base.HoudiniPackageRunner(
                mock_discoverer
            )

        assert inst._discoverer == mock_discoverer
        assert not inst._extra_args
        assert inst._hotl_command == "hotl"
        assert inst._temp_dir == pathlib.Path(mock_mkdtemp.return_value)
        assert not inst._verbose
        assert inst._write_back == mock_write_back

        if pass_optional:
            assert inst._config == mock_config

        else:
            assert inst._config == mock_init_config.return_value

    # Properties

    def test_config(self, mocker, init_runner):
        """Test HoudiniPackageRunner.config."""
        mock_config = mocker.MagicMock(
            spec=houdini_package_runner.config.BaseRunnerConfig
        )

        inst = init_runner()
        inst._config = mock_config

        assert inst.config == mock_config

        with pytest.raises(AttributeError):
            inst.config = None

    def test_discoverer(self, mocker, init_runner):
        """Test HoudiniPackageRunner.discoverer."""
        mock_discoverer = mocker.MagicMock(spec=BaseItemDiscoverer)

        inst = init_runner()
        inst._discoverer = mock_discoverer

        assert inst.discoverer == mock_discoverer

        with pytest.raises(AttributeError):
            inst.discoverer = None

    def test_extra_args(self, mocker, init_runner):
        """Test HoudiniPackageRunner.extra_args."""
        mock_args = mocker.MagicMock(spec=list)

        inst = init_runner()
        inst._extra_args = mock_args

        assert inst.extra_args == mock_args

        with pytest.raises(AttributeError):
            inst.extra_args = []

    def test_hotl_command(self, mocker, init_runner):
        """Test HoudiniPackageRunner.hotl_command."""
        mock_command = mocker.MagicMock(spec=str)

        inst = init_runner()
        inst._hotl_command = mock_command

        assert inst.hotl_command == mock_command

        with pytest.raises(AttributeError):
            inst.hotl_command = None

    def test_temp_dir(self, mocker, init_runner):
        """Test HoudiniPackageRunner.temp_dir."""
        mock_path = mocker.MagicMock(spec=pathlib.Path)

        inst = init_runner()
        inst._temp_dir = mock_path

        assert inst.temp_dir == mock_path

        with pytest.raises(AttributeError):
            inst.temp_dir = None

    def test_verbose(self, mocker, init_runner):
        """Test HoudiniPackageRunner.verbose."""
        mock_verbose = mocker.MagicMock(spec=bool)

        inst = init_runner()
        inst._verbose = mock_verbose

        assert inst.verbose == mock_verbose

        with pytest.raises(AttributeError):
            inst.temp_dir = False

    def test_write_back(self, mocker, init_runner):
        """Test HoudiniPackageRunner.write_back."""
        mock_write_back = mocker.MagicMock(spec=bool)

        inst = init_runner()
        inst._write_back = mock_write_back

        assert inst.write_back == mock_write_back

        with pytest.raises(AttributeError):
            inst.write_back = False

    # Methods

    @pytest.mark.parametrize("has_hotl", (True, False))
    def test_init_args_options(self, mocker, init_runner, has_hotl):
        """Test HoudiniPackageRunner.init_args_options."""
        mock_verbose = mocker.MagicMock(spec=bool)
        mock_list = mocker.MagicMock(spec=bool)
        mock_hotl = mocker.MagicMock(spec=str)

        namespace = argparse.Namespace()
        namespace.verbose = mock_verbose
        namespace.list_items = mock_list

        if has_hotl:
            namespace.hotl_command = mock_hotl

        inst = init_runner()

        inst.init_args_options(namespace, [])

        assert inst._verbose == mock_verbose

        if has_hotl:
            assert inst._hotl_command == mock_hotl

    @pytest.mark.parametrize(
        "list_items, write_back, return_codes, expected",
        (
            (True, False, (0, 0), 0),
            (False, False, (0, 0), 0),
            (False, True, (0, 1), 1),
            (False, True, (1, 0), 1),
        ),
    )
    def test_run(
        self, mocker, init_runner, list_items, write_back, return_codes, expected
    ):
        """Test HoudiniPackageRunner.run."""
        mock_file = mocker.MagicMock(
            spec=houdini_package_runner.items.filesystem.FileToProcess
        )
        mock_file.write_back = False
        mock_file.process.return_value = return_codes[0]

        mock_dir = mocker.MagicMock(
            spec=houdini_package_runner.items.filesystem.DirectoryToProcess
        )
        mock_dir.write_back = False
        mock_dir.process.return_value = return_codes[1]

        mock_discoverer = mocker.MagicMock(spec=BaseItemDiscoverer)
        type(mock_discoverer).items = [mock_file, mock_dir]

        mocker.patch.object(
            houdini_package_runner.runners.base.HoudiniPackageRunner,
            "discoverer",
            mock_discoverer,
        )

        inst = init_runner()
        inst._list_items = list_items
        inst._write_back = write_back

        result = inst.run()
        assert result == expected

        if not list_items:
            mock_file.process.assert_called_with(inst)
            mock_dir.process.assert_called_with(inst)

        if write_back:
            assert mock_file.write_back
            assert mock_dir.write_back



"""This module contains a class to discover package items."""

# =============================================================================
# IMPORTS
# =============================================================================

# Future
from __future__ import annotations

# Standard Library
import pathlib
from typing import TYPE_CHECKING, List, Optional

# Houdini Package Runner
import houdini_package_runner.parser
from houdini_package_runner.discoverers.base import BaseItemDiscoverer
from houdini_package_runner.items import digital_asset, filesystem, xml

# Imports for type checking.
if TYPE_CHECKING:
    import argparse

    from houdini_package_runner.items.base import BaseFileItem, BaseItem


# =============================================================================
# CLASSES
# =============================================================================


class PackageItemDiscoverer(BaseItemDiscoverer):
    """This class is responsible for searching for various items to process.

    :param path: The root path to start searching for Houdini items from.
    :param houdini_items: A list of Houdini related directories to look for.
    :param extra_paths: A list of extra paths to process.
    :param items: Optional list of specific items to check.

    """

    def __init__(
        self,
        path: pathlib.Path,
        houdini_items: List[str],
        extra_paths: Optional[List[pathlib.Path]] = None,
        items: Optional[List[BaseItem]] = None,
    ) -> None:
        super().__init__(path, items=items)

        self.items.extend(get_houdini_items(houdini_items, path))

        if extra_paths is not None:
            for extra_path in extra_paths:
                if extra_path.is_file():
                    self.items.append(filesystem.FileToProcess(extra_path))

                elif extra_path.is_dir():  # pragma: no branch
                    result = process_directory(extra_path)

                    if result is not None:
                        self.items.append(result)


# =============================================================================
# FUNCTIONS
# =============================================================================


def get_digital_asset_items(otl_path: pathlib.Path) -> List[BaseFileItem]:
    """Get a list of shelf items to process.

    :param otl_path: The path to the otls folder.
    :return: A list otl items.

    """
    items: List[BaseFileItem] = []

    for otl_folder in otl_path.iterdir():
        if otl_folder.is_dir() and digital_asset.is_expanded_digital_asset_dir(
            otl_folder
        ):
            items.append(digital_asset.DigitalAssetDirectory(otl_folder))

        elif otl_folder.suffix in (".hda", ".otl"):
            items.append(digital_asset.BinaryDigitalAssetFile(otl_folder))

    return items


def get_houdini_items(
    houdini_items: List[str], houdini_root: pathlib.Path
) -> List[BaseItem]:
    """Get Houdini-related items to process.

    :param houdini_items: The Houdini item names.
    :param houdini_root: The root houdini directory.
    :return: The Houdini items to process.

    """
    items: List[BaseItem] = []

    for item_name in houdini_items:
        if not item_name:
            continue

        item_path = houdini_root / item_name

        if item_name in ("otls", "hda"):
            if not item_path.is_dir():
                continue

            items.extend(get_digital_asset_items(item_path))

        elif item_name == "toolbar":
            items.extend(get_tool_items(item_path))

        elif item_name == "python_panels":
            items.extend(get_python_panel_items(item_path))

        elif item_name == "menus":
            items.extend(get_menu_items(houdini_root))

        elif item_name == "pythonXlibs":
            python_lib_dirs = houdini_root.glob("python*libs")

            for lib_dir in python_lib_dirs:
                items.append(
                    filesystem.HoudiniDirectoryItem(lib_dir, traverse_children=True)
                )

        else:
            if not item_path.is_dir():
                continue

            item = process_directory(item_path)

            if item is not None:
                items.append(item)

    return items


def get_menu_items(houdini_root: pathlib.Path) -> List[xml.MenuFile]:
    """Get a list of xml menu items to process.

    :param houdini_root: The path to a houdini directory.
    :return: A list of menu items.

    """
    menu_files = [xml.MenuFile(menu_file) for menu_file in houdini_root.glob("*.xml")]

    return menu_files


def get_python_panel_items(
    python_panel_path: pathlib.Path,
) -> List[xml.PythonPanelFile]:
    """Get a list of python panel items to process.

    :param python_panel_path: The path to a python_panel directory.
    :return: A list of shelf python panel items.

    """
    panel_files = [
        xml.PythonPanelFile(panel_file)
        for panel_file in python_panel_path.glob("*.pypanel")
    ]

    return panel_files


def get_tool_items(toolbar_path: pathlib.Path) -> List[xml.ShelfFile]:
    """Get a list of shelf items to process.

    :param toolbar_path: The path to a toolbar directory.
    :return: A list of shelf tool items.

    """
    shelf_files = [
        xml.ShelfFile(shelf_file) for shelf_file in toolbar_path.glob("*.shelf")
    ]

    return shelf_files


def init_standard_package_discoverer(
    parsed_args: argparse.Namespace,
) -> PackageItemDiscoverer:
    """Create a standard PackageItemDiscoverer based on standard args.

    :param parsed_args: The parsed script args.
    :return: A discoverer object based on the parsed args.

    """
    (
        _,
        houdini_path,
        extra_paths,
        houdini_items,
    ) = houdini_package_runner.parser.process_common_arg_settings(parsed_args)

    discoverer = PackageItemDiscoverer(
        houdini_path,
        houdini_items=houdini_items,
        extra_paths=extra_paths,
    )

    return discoverer


def process_directory(dir_path: pathlib.Path) -> filesystem.DirectoryToProcess:
    """Process a directory to determine its item type.

    :param dir_path: The directory to process.
    :return: The corresponding directory item type.

    """
    if (dir_path / "__init__.py").exists():
        return filesystem.PythonPackageDirectoryItem(dir_path)

    if dir_path.name == "python":
        return filesystem.PythonPackageDirectoryItem(dir_path)

    if dir_path.name == "scripts":
        return filesystem.HoudiniScriptsDirectoryItem(dir_path, traverse_children=True)

    item = filesystem.DirectoryToProcess(dir_path, traverse_children=True)

    if dir_path.name == "tests":
        item.is_test_item = True

    return item



"""This module contains a base runnable item."""

# =============================================================================
# IMPORTS
# =============================================================================

# Future
from __future__ import annotations

# Standard Library
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, List

# Imports for type checking.
if TYPE_CHECKING:
    import pathlib

    import houdini_package_runner.runners.base


# =============================================================================
# CLASSES
# =============================================================================


class BaseItem(ABC):
    """Base class for a runnable item.

    :param write_back: Whether the item should write itself back to disk.

    """

    def __init__(self, write_back: bool = False) -> None:
        self._contents_changed = False
        self._ignored_builtins: List[str] = []
        self._is_single_line = False
        self._is_test_item = False
        self._write_back = write_back

    def __repr__(self):
        return f"<{self.__class__.__name__}>"

    # -------------------------------------------------------------------------
    # PROPERTIES
    # -------------------------------------------------------------------------

    @property
    def contents_changed(self) -> bool:
        """Whether the contents of the item have changed."""
        return self._contents_changed

    @contents_changed.setter
    def contents_changed(self, contents_changed: bool):
        self._contents_changed = contents_changed

    # -------------------------------------------------------------------------

    @property
    def ignored_builtins(self) -> List[str]:
        """A list of known builtins to ignore for checks which look for imports."""
        return self._ignored_builtins

    # -------------------------------------------------------------------------

    @property
    def is_single_line(self) -> bool:
        """Whether the item code on a single line."""
        return self._is_single_line

    # -------------------------------------------------------------------------

    @property
    def is_test_item(self) -> bool:
        """Whether the item is a test related item."""
        return self._is_test_item

    @is_test_item.setter
    def is_test_item(self, is_test_item: bool):
        self._is_test_item = is_test_item

    # -------------------------------------------------------------------------

    @property
    def write_back(self) -> bool:
        """Whether the item should write changes back."""
        return self._write_back

    @write_back.setter
    def write_back(self, write_back):
        self._write_back = write_back

    # -------------------------------------------------------------------------
    # METHODS
    # -------------------------------------------------------------------------

    @abstractmethod
    def process(
        self, runner: houdini_package_runner.runners.base.HoudiniPackageRunner
    ) -> int:
        """Process an item.

        :param runner: The package runner processing the item.
        :return: The process return code.

        """


class BaseFileItem(BaseItem):
    """Base class for a runnable item.

    :param path: The path for the item.
    :param write_back: Whether the item should write itself back to disk.

    """

    def __init__(self, path: pathlib.Path, write_back: bool = False) -> None:
        super().__init__(write_back=write_back)
        self._path = path

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.path}>"

    # -------------------------------------------------------------------------
    # PROPERTIES
    # -------------------------------------------------------------------------

    @property
    def path(self) -> pathlib.Path:
        """The path on disk."""
        return self._path

    # -------------------------------------------------------------------------
    # METHODS
    # -------------------------------------------------------------------------

    @abstractmethod
    def process(
        self, runner: houdini_package_runner.runners.base.HoudiniPackageRunner
    ) -> int:
        """Process an item.

        :param runner: The package runner processing the item.
        :return: The process return code.

        """



"""This module contains runnable items based on Houdini digital asset's DialogScript."""

# =============================================================================
# IMPORTS
# =============================================================================

# Future
from __future__ import annotations

# Standard Library
import abc
import operator
from collections import namedtuple
from typing import TYPE_CHECKING, List, Optional, Tuple

# Third Party
import pyparsing  # type: ignore

# Houdini Package Runner
from houdini_package_runner.items.base import BaseFileItem, BaseItem

# Imports for type checking.
if TYPE_CHECKING:
    import pathlib

    import houdini_package_runner.runners.base


# =============================================================================
# GLOBALS
# =============================================================================

# The majority of the code below to do the parsing of the DialogScript file
# is lifted from Houdini's $HFS/bin/py23convert.py tool.  A big thanks to them
# for the unenviable task of writing this such that we may repurpose it for
# our own use.

_DS_NAME_EXPR_GRAMMAR = (
    pyparsing.Keyword("name").suppress() + pyparsing.QuotedString('"')
).parseWithTabs()

_DS_DEF_EXPR_TYPED = (
    pyparsing.Literal("[").suppress()
    + pyparsing.Group(
        pyparsing.locatedExpr(
            pyparsing.QuotedString(quoteChar='"', escChar="\\", multiline=True)
        )
        + pyparsing.Word(pyparsing.alphas + "-")
    )
    + pyparsing.Literal("]").suppress()
)

_DS_DEF_EXPR_LIT = (
    pyparsing.QuotedString(quoteChar='"', escChar="\\", multiline=True).suppress()
    | pyparsing.Word(pyparsing.nums).suppress()
)

_DS_DEF_EXPR_GRAMMAR = (
    pyparsing.Keyword("default").suppress()
    + pyparsing.Literal("{").suppress()
    + pyparsing.OneOrMore(_DS_DEF_EXPR_TYPED | _DS_DEF_EXPR_LIT)
    + pyparsing.Literal("}").suppress()
).parseWithTabs()

_DS_CB_SCRIPT_LANG_GRAMMAR = (
    pyparsing.Keyword("parmtag").suppress()
    + pyparsing.Literal("{").suppress()
    + pyparsing.Literal('"script_callback_language"').suppress()
    + pyparsing.QuotedString('"')
    + pyparsing.Literal("}").suppress()
)

_DS_CB_SCRIPT_GRAMMAR = (
    pyparsing.Keyword("parmtag").suppress()
    + pyparsing.Literal("{").suppress()
    + pyparsing.Literal('"script_callback"').suppress()
    + pyparsing.locatedExpr(
        pyparsing.QuotedString(quoteChar='"', escChar="\\", multiline=True)
    )
    + pyparsing.Literal("}").suppress()
).parseWithTabs()

_DS_MENU_PY = pyparsing.locatedExpr(
    pyparsing.Literal("[").suppress()
    + pyparsing.QuotedString(quoteChar='"', escChar="\\")
    + pyparsing.Literal("]").suppress()
).parseWithTabs()

_DS_MENU_GRAMMAR = (
    pyparsing.Keyword("menu").suppress()
    + pyparsing.Literal("{").suppress()
    + pyparsing.OneOrMore(_DS_MENU_PY)
    + pyparsing.Literal("language").suppress()
    + pyparsing.Literal("python").suppress()
    + pyparsing.Literal("}").suppress()
).parseWithTabs()

_DS_PARM_GRAMMAR = pyparsing.originalTextFor(
    pyparsing.Keyword("parm") + pyparsing.nestedExpr("{", "}")
).parseWithTabs()


PythonMenuScriptResult = namedtuple(
    "PythonMenuScriptResult", ["menu_script", "span", "indent", "uses_tabs"]
)
"""Item to hold menu script information."""


# =============================================================================
# CLASSES
# =============================================================================


class DialogScriptItem(BaseFileItem):
    """Item representing a DialogScript section inside a digital asset definition.

    :param path: The file path to process.
    :param name: The display name for test output.
    :param write_back: Whether the item should write itself back to disk.
    :param source_file: Optional source file for the expanded operator definition.

    """

    def __init__(
        self,
        path: pathlib.Path,
        name: str,
        write_back: bool = False,
        source_file: pathlib.PurePath = None,
    ) -> None:
        super().__init__(path, write_back=write_back)

        self._name = name
        self._source_file = source_file

        # Stash the script contents.
        with self.path.open(encoding="utf-8") as handle:
            self._ds_contents = handle.read()

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.name} ({self.path})>"

    # -------------------------------------------------------------------------
    # NON-PUBLIC METHODS
    # -------------------------------------------------------------------------

    def _gather_items(self) -> Tuple[DialogScriptInternalItem, ...]:
        """Gather the items from the DialogScript to process.

        :return: Any items to be processed.

        """
        items: List[DialogScriptInternalItem] = []

        for match, parm_start, _ in _DS_PARM_GRAMMAR.scanString(self._ds_contents):
            # The block of code related to the parameter definition.
            parm = match.asList()[0]

            items.extend(_get_callback_items(parm, parm_start, self.name))

            items.extend(_get_expression_items(parm, parm_start, self.name))

            items.extend(_get_menu_items(parm, parm_start, self.name))

        # Set the 'write_back' property of all the items if necessary.
        if self.write_back:
            for item in items:
                item.write_back = True

        return tuple(items)

    def _handle_changed_contents(
        self, items_with_changed_contents: List[DialogScriptInternalItem]
    ):
        """Handle writing any items with changed contents to the file.

        :param items_with_changed_contents: A list of items with changed contents.

        """
        items_with_changed_contents.sort(key=operator.attrgetter("start_offset"))

        parts = []

        script_offset = 0

        for item in items_with_changed_contents:
            parts.append(self._ds_contents[script_offset : item.start_offset])
            parts.append(item.post_processed_code)
            script_offset = item.end_offset

        parts.append(self._ds_contents[script_offset:])
        new_dialog_script = "".join(parts)

        with open(self.path, "w", encoding="utf-8") as handle:
            handle.write(new_dialog_script)

    # -------------------------------------------------------------------------
    # PROPERTIES
    # -------------------------------------------------------------------------

    @property
    def name(self) -> str:
        """The display name of the operator."""
        return self._name

    # -------------------------------------------------------------------------
    # METHODS
    # -------------------------------------------------------------------------

    def process(
        self, runner: houdini_package_runner.runners.base.HoudiniPackageRunner
    ) -> int:
        """Process the operator files.

        :param runner: The package runner processing the item.
        :return: The process return code.

        """
        files_to_process = self._gather_items()

        result = 0

        items_with_changed_contents = []

        for file_to_process in files_to_process:
            result |= file_to_process.process(runner)

            if file_to_process.contents_changed:
                items_with_changed_contents.append(file_to_process)

        if self.write_back and items_with_changed_contents:
            self.contents_changed = True
            self._handle_changed_contents(items_with_changed_contents)

        return result


class DialogScriptInternalItem(BaseItem, metaclass=abc.ABCMeta):
    """Base definition for a Python processable item in a DialogScript.

    :param parm: The source parameter definition.
    :param code: The Python code for the item.
    :param start_offset: There parameter definition start offset.
    :param end_offset: There parameter definition end offset.
    :param display_name: The item display name.
    :param write_back: Whether the item should write itself back to disk.

    """

    def __init__(
        self,
        parm: str,
        code: str,
        start_offset: int,
        end_offset: int,
        display_name: str,
        write_back: bool = False,
    ) -> None:
        super().__init__(write_back=write_back)

        self._code = code
        self._end_offset = end_offset
        self._parm = parm
        self._start_offset = start_offset

        name = _DS_NAME_EXPR_GRAMMAR.searchString(parm)[0][0]

        self._name = name
        self._display_hint = ""
        self._display_name = display_name + "_" + name

        self._post_processed_code = code

        # The 'hou' module is always available in these items.
        self.ignored_builtins.append("hou")

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.display_name}>"

    # -------------------------------------------------------------------------
    # NON-PUBLIC METHODS
    # -------------------------------------------------------------------------

    def _load_contents(self, temp_path: pathlib.Path) -> str:
        """Load the contents from the temp path.

        :param temp_path: The item file path.
        :return: The file contents.

        """
        with temp_path.open("r") as handle:
            contents = handle.read()

        # Strip any final newline that we probably added above.
        if contents.endswith("\n"):
            contents = contents[:-1]

        if self.is_single_line:
            contents = _escape_contents_for_single_line(contents)

        return contents

    def _post_process_contents(  # pylint: disable=no-self-use
        self, contents: str
    ) -> str:
        """Perform any post-processing on the contents.

        :param contents: The script contents.
        :return: The processed contents.

        """
        return contents

    def _write_contents(self, temp_path: pathlib.Path) -> None:
        """Write the contents of the item to the temp path.

        :param temp_path: The item file path.
        :return:

        """
        code = self.code

        # If the code blob does not end with a new line it can cause problems with
        # runners such as black.  Add a new line at the end of the code to more properly
        # resemble an actual file to be processed.
        if not code.endswith("\n"):
            code += "\n"

        # Dump the code to the temp file, so it can be processed.
        with temp_path.open("w") as handle:
            handle.write(code)

    # -------------------------------------------------------------------------
    # PROPERTIES
    # -------------------------------------------------------------------------

    @property
    def code(self) -> str:
        """The Python code for the item."""
        return self._code

    @property
    def display_name(self) -> str:
        """The item display name."""
        display_name = self._display_name

        if self._display_hint:
            display_name += "__" + self._display_hint

        return display_name

    @property
    def end_offset(self) -> int:
        """There parameter definition end offset."""
        return self._end_offset

    @property
    def name(self) -> str:
        """The parameter name."""
        return self._name

    @property
    def parm(self) -> str:
        """The parameter definition."""
        return self._parm

    @property
    def start_offset(self) -> int:
        """There parameter definition start offset."""
        return self._start_offset

    @property
    def post_processed_code(self) -> str:
        """The post-processing contents"""
        return self._post_processed_code

    # -------------------------------------------------------------------------
    # METHODS
    # -------------------------------------------------------------------------

    def process(
        self, runner: houdini_package_runner.runners.base.HoudiniPackageRunner
    ) -> int:
        """Process an item.

        :param runner: The package runner processing the item.
        :return: The process return code.

        """
        temp_path = runner.temp_dir / f"{self.display_name}.py"

        self._write_contents(temp_path)

        result = runner.process_path(temp_path, self)

        if self.write_back:
            contents = self._load_contents(temp_path)

            if self.code != contents:
                self.contents_changed = True

                self._post_processed_code = self._post_process_contents(contents)

        return result


class DialogScriptCallbackItem(DialogScriptInternalItem):
    """Item to represent and process a parameter's Python callback script.

    :param parm: The source parameter definition.
    :param code: The Python code for the item.
    :param parm_start: The start position of the parm.
    :param span: The span of the parm data.
    :param display_name: The item display name.
    :param write_back: Whether the item should write itself back to disk.

    """

    def __init__(
        self,
        parm: str,
        code: str,
        parm_start: int,
        span: Tuple[int, int],
        display_name: str,
        write_back: bool = False,
    ) -> None:
        start_offset, end_offset = _get_ds_file_offset(parm_start, span)

        super().__init__(
            parm, code, start_offset, end_offset, display_name, write_back=write_back
        )

        self._display_hint = "callback"
        self._is_single_line = True

        # kwargs is always available in callback scripts.
        self.ignored_builtins.append("kwargs")


class DialogScriptDefaultExpressionItem(DialogScriptInternalItem):
    """Item to represent and process a parameter's Python default expression.

    :param parm: The source parameter definition.
    :param code: The Python code for the item.
    :param parm_start: The start position of the parm.
    :param span: The span of the parm data.
    :param display_name: The item display name.
    :param write_back: Whether the item should write itself back to disk.

    """

    def __init__(
        self,
        parm: str,
        code: str,
        parm_start: int,
        span: Tuple[int, int],
        display_name: str,
        write_back: bool = False,
    ) -> None:
        start_offset, end_offset = _get_ds_file_offset(parm_start, span)

        super().__init__(
            parm, code, start_offset, end_offset, display_name, write_back=write_back
        )

        self._display_hint = "default"
        self._is_single_line = True


class DialogScriptMenuScriptItem(DialogScriptInternalItem):
    """Item to represent and process a parameter's Python menu script.

    :param parm: The source parameter definition.
    :param parm_start: The start position of the parm.
    :param display_name: The item display name.
    :param menu_script_data: The Python menu script data
    :param write_back: Whether the item should write itself back to disk.

    """

    def __init__(
        self,
        parm: str,
        parm_start: int,
        display_name: str,
        menu_script_data: PythonMenuScriptResult,
        write_back: bool = False,
    ) -> None:
        start_offset, end_offset = _get_ds_file_offset(
            parm_start, menu_script_data.span, inclusive=True
        )

        super().__init__(
            parm,
            menu_script_data.menu_script,
            start_offset,
            end_offset,
            display_name,
            write_back=write_back,
        )

        self._display_hint = "menu_script"
        self._menu_script_data = menu_script_data

        # kwargs is always available in menu scripts.
        self.ignored_builtins.append("kwargs")

    # -------------------------------------------------------------------------
    # NON-PUBLIC METHODS
    # -------------------------------------------------------------------------

    def _post_process_contents(self, contents: str) -> str:
        """Perform any post-processing on the contents.

        This will ensure the multiline menu script is properly wrapped.

        :param contents: The script contents.
        :return: The processed contents.

        """
        lines = contents.splitlines()
        new_contents = ""

        indent_char = "\t" if self.menu_script_data.uses_tabs else " "
        indent = indent_char * self.menu_script_data.indent

        for i, line in enumerate(lines):
            new_contents += f'{"" if i == 0 else indent}[ "{line}" ]\n'

        return new_contents

    # -------------------------------------------------------------------------
    # PROPERTIES
    # -------------------------------------------------------------------------

    @property
    def menu_script_data(self) -> PythonMenuScriptResult:
        """The Python menu script data."""
        return self._menu_script_data


# =============================================================================
# NON-PUBLIC FUNCTIONS
# =============================================================================


def _discard_newlines(parm: str, start: int) -> int:
    """Discard any newline characters.

    :param parm: The parameter data.
    :param start: The start index.
    :return: The start index offset to discard any newlines

    """
    pos = start

    while pos < len(parm):
        if parm[pos] not in ("\r", "\n"):
            break

        pos += 1

    return pos


def _escape_contents_for_single_line(contents: str) -> str:
    """Escape characters to write as a single line.

    :param contents: The contents to escape.
    :return: The contents with \r, \n and " escaped

    """
    contents = contents.replace("\r", "\\r").replace("\n", "\\n").replace('"', '\\"')

    return contents


def _get_callback_items(
    parm: str, parm_start: int, name: str
) -> List[DialogScriptCallbackItem]:
    """Build a list of any callback items for a parameter.

    :param parm: The parameter data.
    :param parm_start: The start position of the parm.
    :param name: The display name.
    :return: A list of any callback items.

    """
    items = []
    # Get the parameter's callback script language.
    callback_language = _get_callback_language(parm)

    if callback_language == "python":
        # Get the callback script data, if any.
        result = _get_script_callback(parm)

        if result is not None:
            script, script_span = result

            items.append(
                DialogScriptCallbackItem(parm, script, parm_start, script_span, name)
            )

    return items


def _get_callback_language(parm: str) -> Optional[str]:
    """Get a parameter's callback script language.

    :param parm: The parameter data.
    :return: The script callback language.

    """
    for match in _DS_CB_SCRIPT_LANG_GRAMMAR.searchString(parm):
        return match[0]

    return None


def _get_ds_file_offset(
    parm_start: int, span: Tuple[int, int], inclusive: bool = False
) -> Tuple[int, int]:
    """Get the file offsets for the parameter.

    :param parm_start: The start position of the parm.
    :param span: The span of the parm data.
    :param inclusive: Whether to include the start and end chars.
    :return: The file offsets for the parameter.

    """
    extra_offset = 0 if inclusive else 1

    return parm_start + span[0] + extra_offset, parm_start + span[1] - extra_offset


def _get_default_python_expressions(
    parm: str,
) -> Tuple[Tuple[str, Tuple[int, int]], ...]:
    """Get default Python expressions for a parameter tuple.

    :param parm: The parameter data.
    :return: Any default expressions which are Python.

    """
    expressions = []

    for matches in _DS_DEF_EXPR_GRAMMAR.searchString(parm):
        for match in matches:
            (start, expr, end), expr_lang = match

            if expr_lang == "python":
                expressions.append((expr, (start, end)))

    return tuple(expressions)


def _get_expression_items(
    parm: str, parm_start: int, name: str
) -> List[DialogScriptDefaultExpressionItem]:
    """Build a list of any expression items for a parameter.

    :param parm: The parameter data.
    :param parm_start: The start position of the parm.
    :param name: The display name.
    :return: A list of any expression items.

    """
    items = []

    default_python_expressions = _get_default_python_expressions(parm)

    for expr, span in default_python_expressions:
        items.append(
            DialogScriptDefaultExpressionItem(parm, expr, parm_start, span, name)
        )

    return items


def _get_menu_items(
    parm: str, parm_start: int, name: str
) -> List[DialogScriptMenuScriptItem]:
    """Build a list of any menu script items for a parameter.

    :param parm: The parameter data.
    :param parm_start: The start position of the parm.
    :param name: The display name.
    :return: A list of any menu script items.

    """
    items = []

    # Get any Python menu script data if it exists.
    menu_script = _get_python_menu_script(parm)

    if menu_script is not None:
        items.append(DialogScriptMenuScriptItem(parm, parm_start, name, menu_script))

    return items


def _get_python_menu_script(parm: str) -> Optional[PythonMenuScriptResult]:
    """Get any Python menu script data for a parameter.

    :param parm: The parameter data.
    :return: The menu script data if the parameter has a Python one.

    """
    for match in _DS_MENU_GRAMMAR.searchString(parm):
        lines = match.asList()
        start = lines[0][0]
        end = _discard_newlines(parm, lines[-1][2])
        menu_script = "\n".join([line[1] for line in lines])
        indent = 0

        uses_tabs = False

        for i in range(start - 1, 0, -1):  # pragma: no branch
            if parm[i] == " ":
                indent += 1

            elif parm[i] == "\t":
                indent += 1
                uses_tabs = True

            else:
                break

        # There can only be one menu script so return it.
        return PythonMenuScriptResult(menu_script, (start, end), indent, uses_tabs)

    return None


def _get_script_callback(parm: str) -> Optional[Tuple[str, Tuple[int, int]]]:
    """Get the callback script data for a parameter if it has one.

    The result contains the script contents and start/end offset values.

    :param parm: The parameter data.
    :return: The parameter callback script if the parameter has one.

    """
    for match in _DS_CB_SCRIPT_GRAMMAR.searchString(parm):
        # There can only been one callback script entry per parameter.
        start, text, end = match[0]
        return text, (start, end)

    return None



"""Test the houdini_package_runner.parser module."""

# =============================================================================
# IMPORTS
# =============================================================================

# Standard Library
import argparse
import pathlib

# Third Party
import pytest

# Houdini Package Runner
import houdini_package_runner.parser

# =============================================================================
# TESTS
# =============================================================================


@pytest.mark.parametrize(
    "extra_path, root_value, use_abs_path, raises",
    (
        (None, "houdini", True, False),
        (None, "houdini", False, False),
        (None, "other", False, True),
        ("houdini", None, False, False),
        (None, None, False, False),
    ),
)
def test__resolve_houdini_root(
    shared_datadir, extra_path, root_value, use_abs_path, raises
):
    """Test houdini_package_runner.parser._resolve_houdini_root."""
    test_root_path = shared_datadir / "resolve_houdini_root"

    abs_value = test_root_path / "houdini"

    if extra_path:
        test_root_path /= extra_path

    if use_abs_path:
        root_value = abs_value

    mock_namespace = argparse.Namespace()
    mock_namespace.houdini_root = root_value if root_value else None

    if raises:
        with pytest.raises(OSError):
            houdini_package_runner.parser._resolve_houdini_root(
                mock_namespace, test_root_path
            )

    else:
        result = houdini_package_runner.parser._resolve_houdini_root(
            mock_namespace, test_root_path
        )

        assert result == shared_datadir / "resolve_houdini_root" / "houdini"


@pytest.mark.parametrize("python_root", (None, "python", "other"))
def test__resolve_python_packages(shared_datadir, python_root):
    """Test houdini_package_runner.parser._resolve_python_packages."""
    test_root_path = shared_datadir / "resolve_python_packages"

    abs_value = test_root_path / "python"

    mock_namespace = argparse.Namespace()
    mock_namespace.python_root = python_root if python_root else None

    result = houdini_package_runner.parser._resolve_python_packages(
        mock_namespace, test_root_path
    )

    if python_root == "python":
        assert result == [abs_value]

    else:
        assert not result


@pytest.mark.parametrize(
    "extra_path, skip_tests",
    (
        (None, True),
        (None, False),
        ("foo", False),
    ),
)
def test__resolve_tests(shared_datadir, extra_path, skip_tests):
    """Test houdini_package_runner.parser._resolve_tests."""
    test_root_path = shared_datadir / "resolve_tests"

    if extra_path:
        test_root_path /= extra_path

    abs_value = test_root_path / "tests"

    mock_namespace = argparse.Namespace()
    mock_namespace.skip_tests = skip_tests

    result = houdini_package_runner.parser._resolve_tests(
        mock_namespace, test_root_path
    )

    if not skip_tests and not extra_path:
        assert result == [abs_value]

    else:
        assert not result


@pytest.mark.parametrize("pass_args", (True, False))
def test_build_common_parser(mocker, pass_args):
    """Test houdini_package_runner.parser.build_common_parser."""
    # Store a reference to the parser class, so we can compare to it later on after
    # we do all sorts of mocking and patching.
    argparse_type = argparse.ArgumentParser

    mock_parser = mocker.MagicMock(spec=argparse.ArgumentParser)
    mock_init_parser = mocker.patch("argparse.ArgumentParser", return_value=mock_parser)

    if pass_args:
        result = houdini_package_runner.parser.build_common_parser(
            prog="program name", usage="usage", description="description"
        )
    else:
        result = houdini_package_runner.parser.build_common_parser()

    assert isinstance(result, argparse_type)

    calls = [
        mocker.call(
            "--add-dir",
            action="append",
            dest="directories",
            default=[],
            help="Add a directory to be processed",
        ),
        mocker.call(
            "--add-file",
            action="append",
            dest="files",
            default=[],
            help="Add a file to be processed",
        ),
        mocker.call(
            "--houdini-items",
            action="store",
            default="otls,hda,toolbar,python_panels,pythonXlibs,scripts,soho,menus",
            help="A list of Houdini items to process",
        ),
        mocker.call(
            "--python-root",
            action="store",
            default="python",
            help="The root directory the package will be in",
        ),
        mocker.call(
            "--root",
            action="store",
            help="Optional root directory to look for things from.  By default uses the cwd",
        ),
        mocker.call("--verbose", action="store_true", help="Engage verbose output."),
        mocker.call("--hotl-command", action="store", default="hotl"),
        mocker.call(
            "--skip-tests",
            action="store_true",
            help="Skip processing of files under {root}/tests.",
        ),
    ]

    if pass_args:
        mock_init_parser.assert_called_with(
            prog="program name",
            usage="usage",
            description="description",
            formatter_class=houdini_package_runner.parser._UltimateHelpFormatter,
        )

    assert mock_parser.add_argument.has_calls(calls)


@pytest.mark.parametrize("root_passed", (True, False))
def test_process_common_arg_settings(mocker, root_passed):
    """Test houdini_package_runner.parser.process_common_arg_settings."""
    namespace = argparse.Namespace()
    namespace.root = pathlib.Path("/some/root") if root_passed else None
    namespace.directories = ["directory1"]
    namespace.files = ["file1"]

    mock_python_dir = mocker.MagicMock(spec=pathlib.Path)
    mock_test_dir = mocker.MagicMock(spec=pathlib.Path)

    mock_houdini_root = mocker.patch(
        "houdini_package_runner.parser._resolve_houdini_root"
    )
    mocker.patch(
        "houdini_package_runner.parser._resolve_python_packages",
        return_value=[mock_python_dir],
    )
    mocker.patch(
        "houdini_package_runner.parser._resolve_tests", return_value=[mock_test_dir]
    )

    namespace.houdini_items = "otls,hda,python3.7libs"

    expected_root = namespace.root if root_passed else pathlib.Path.cwd()

    expected_extra = [
        expected_root / "directory1",
        mock_python_dir,
        mock_test_dir,
        expected_root / "file1",
    ]

    result = houdini_package_runner.parser.process_common_arg_settings(namespace)

    assert result == (
        expected_root,
        mock_houdini_root.return_value,
        expected_extra,
        namespace.houdini_items.split(","),
    )
